"""
Utility functions for causalexplain
(C) J. Renero, 2022, 2023, 2024
"""

# pylint: disable=E1101:no-member, W0201:attribute-defined-outside-init
# pylint: disable=C0103:invalid-name, W0511:fixme
# pylint: disable=C0116:missing-function-docstring
# pylint: disable=R0913:too-many-arguments
# pylint: disable=R0914:too-many-locals, R0915:too-many-statements
# pylint: disable=W0106:expression-not-assigned, R1702:too-many-branches

import glob
import json
import os
import pickle
import types
from os.path import join
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import networkx as nx
import numpy as np
import pandas as pd
import pydot
import pydotplus
import torch

AnyGraph = Union[nx.Graph, nx.DiGraph]


def save_experiment(
        obj_name: str,
        folder: str,
        results: object,
        overwrite: bool = False) -> str:
    """
    Creates a folder for the experiment and saves results. Results is a
    dictionary that will be saved as an opaque pickle. When the experiment will
    require to be loaded, the only parameter needed are the folder name.

    Args:
        obj_name (str): the name to be given to the pickle file to be saved. If
            a file already exists with that name, a file with same name and a
            extension will be generated.
        folder (str): a full path to the folder where the experiment is to
            be saved. If the folder does not exist it will be created.
        results (obj): the object to be saved as experiment. This is typically a
            dictionary with different items representing different parts of the
            experiment.
        overwrite (bool): If True, then the experiment is saved even if a file
            with the same name already exists. If False, then the experiment is
            saved only if a file with the same name does not exist.

    Return:
        (str) The name under which the experiment has been saved
    """
    if not os.path.exists(folder):
        Path(folder).mkdir(parents=False, exist_ok=True)

    if not overwrite:
        output = valid_output_name(obj_name, folder, extension="pickle")
    else:
        output = join(folder, obj_name if obj_name.endswith('.pickle') \
            else obj_name + '.pickle')

    with open(output, 'wb') as handle:
        pickle.dump(results, handle, protocol=pickle.HIGHEST_PROTOCOL)

    return output


def load_experiment(obj_name: str, folder: str):
    """
    Loads a pickle from a given folder name. It is not necessary to add the "pickle"
    extension to the experiment name.

    Parameters:
    -----------
    obj_name (str): The name of the object saved in pickle format that is to be loaded.
    folder (str): A full path where looking for the experiment object.

    Returns:
    --------
    An obj loaded from a pickle file.
    """
    if Path(obj_name).suffix == "" or Path(obj_name).suffix != "pickle":
        ext = ".pickle"
    else:
        ext = ''

    experiment = f"{str(Path(folder, obj_name))}{ext}"
    with open(experiment, 'rb') as h:
        obj = pickle.load(h, encoding="utf-8")
    return obj


def valid_output_name(filename: str, path: str, extension=None) -> str:
    """
    Builds a valid name. In case there's another file which already exists
    adds a number (1, 2, ...) until finds a valid filename which does not
    exist.

    Returns
    -------
    The filename if the name is valid and file does not exists,
            None otherwise.

    Params
    ------
    filename: str
        The base filename to be set.
    path: str
        The path where trying to set the filename
    extension:str
        The extension of the file, without the dot '.' If no extension is
        specified, any extension is searched to avoid returning a filepath
        of an existing file, no matter what extension it has.
    """
    if extension:
        if extension[0] == '.':
            extension = extension[1:]
        if filename.endswith(f".{extension}"):
            base_filepath = join(path, filename)
        else:
            base_filepath = join(path, filename) + '.{}'.format(extension)
    else:
        base_filepath = join(path, filename)
    output_filepath = base_filepath
    idx = 1
    while len(glob.glob(f"{output_filepath}*")) > 0:
        if extension:
            output_filepath = join(
                path, filename) + '_{:d}.{}'.format(
                idx, extension)
        else:
            output_filepath = join(path, filename + '_{}'.format(idx))
        idx += 1

    return output_filepath


def graph_from_dot_file(dot_file: Union[str, Path]) -> Optional[nx.DiGraph]:
    """ Returns a NetworkX DiGraph from a DOT FILE. """
    # Check if "dot_file" exists
    if not os.path.exists(dot_file):
        return None
    try:
        dot_object = pydot.graph_from_dot_file(dot_file)
        if dot_object is None or not dot_object or not isinstance(dot_object, (list, tuple)):
            return None

        dot_string = dot_object[0].to_string()
        if not dot_string:
            return None

        dot_string = dot_string.replace('\"\\n\"', '')
        dot_string = dot_string.replace("\n;\n", "\n")
        dotplus = pydotplus.graph_from_dot_data(dot_string)
        if dotplus is None:
            return None

        dotplus.set_strict(True)   # type: ignore
        final_graph = nx.nx_pydot.from_pydot(dotplus)
        if '\\n' in final_graph.nodes:
            final_graph.remove_node('\\n')

        return final_graph
    except Exception as e:
        print(f"Error processing dot file: {str(e)}")
        return None


def graph_from_dictionary(d: Dict[str, List[Union[str, Tuple[str, float]]]]) -> AnyGraph:
    """
    Builds a graph from a dictionary like {'u': ['v', 'w'], 'x': ['y']}.
    The elements of the list can be tuples including weight

    Args:
        d (dict): A dictionary of the form {'u': ['v', 'w'], 'x': ['y']}, where
            an edge between 'u' goes towards 'v' and 'w', and also an edge from
            'x' goes towards 'y'. The format can also be like:
            {'u': [('v', 0.2), ('w', 0.7)], 'x': [('y', 0.5)]}, where the values
            in the tuple are interpreted as weights.

    Returns:
        networkx.DiGraph with the nodes and edges specified.
    """
    g = nx.DiGraph()
    for node, parents in d.items():
        if len(parents) > 0:
            if isinstance(parents[0], tuple):
                for parent, weight in parents:
                    g.add_edge(parent, node, weight=weight)
            else:
                for parent in parents:
                    g.add_edge(parent, node)
    return g


def graph_from_adjacency(
        adjacency: np.ndarray,
        node_labels=None,
        th=0.0,
        inverse: bool = False,
        absolute_values: bool = False
) -> nx.DiGraph:
    """
    Manually parse the adj matrix to shape a dot graph

    Args:
        adjacency: a numpy adjacency matrix
        node_labels: an array of same length as nr of columns in the adjacency
            matrix containing the labels to use with every node.
        th: (float) weight threshold to be considered a valid edge.
        inverse (bool): Set to true if rows in adjacency reflects where edges
            are comming from, instead of where are they going to.
        absolute_values: Take absolute value of weight label to check if
            its greater than the threshold.

    Returns:
         The Graph (DiGraph)
    """
    G = nx.DiGraph()
    G.add_nodes_from(range(adjacency.shape[1]))

    # What to do with absolute values?
    def not_abs(x):
        return x

    w_val = np.abs if absolute_values else not_abs

    def weight_gt(w, thresh):
        return w != 0.0 if thresh is None else w_val(w) > thresh

    # Do I have a threshold to consider?
    for i, row in enumerate(adjacency):
        for j, value in enumerate(row):
            if inverse:
                if weight_gt(adjacency[j][i], th):
                    G.add_edge(i, j, weight=w_val(adjacency[j][i]))
            else:
                if weight_gt(value, th):
                    # , arrowhead="normal")
                    G.add_edge(i, j, weight=w_val(value))
    # Map the current column numbers to the letters used in toy dataset
    if node_labels is not None and len(node_labels) == adjacency.shape[1]:
        mapping = dict(zip(sorted(G), node_labels))
        G = nx.relabel_nodes(G, mapping)

    return G


def graph_from_adjacency_file(
        file: Union[Path, str],
        labels=None,
        th=0.0,
        sep=",",
        header:bool=True
) -> Tuple[nx.DiGraph, pd.DataFrame]:
    """
    Read Adjacency matrix from a file and return a Graph

    Args:
        file: (str) the full path of the file to read
        labels: (List[str]) the list of node names to be used. If None, the
            node names are extracted from the adj file. The names must be
            already sorted in the same order as the adjacency matrix.
        th: (float) weight threshold to be considered a valid edge.
        sep: (str) the separator used in the file
        header: (bool) whether the file has a header. If True, then 'infer'
        is used to read the header.
    Returns:
        DiGraph, DataFrame
    """
    header_infer = "infer" if header else None
    df = pd.read_csv(file, dtype="str", delimiter=sep, header=header_infer)
    df = df.astype("float64")
    if labels is None:
        labels = list(df.columns)
    G = graph_from_adjacency(df.values, node_labels=labels, th=th)

    return G, df


def graph_to_adjacency(
        graph: AnyGraph,
        labels: List[str],
        weight_label: str = "weight") -> np.ndarray:
    """
    A method to generate the adjacency matrix of the graph. Labels are
    sorted for better readability.

    Args:
        graph: (Union[Graph, DiGraph]) the graph to be converted.
        node_names: (List[str]) the list of node names to be used. If None, the
            node names are extracted from the graph. The names must be already
            sorted in the same order as the adjacency matrix.
        weight_label: the label used to identify the weights.

    Return:
        graph: (numpy.ndarray) A 2d array containing the adjacency matrix of
            the graph.
    """
    symbol_map = {"o": 1, ">": 2, "-": 3}

    # Fix for the case where an empty node is parsed from the .dot file
    if '\\n' in labels:
        labels.remove('\\n')

    mat = np.zeros((len(labels), (len(labels))))
    for x in labels:
        for y in labels:
            if graph.has_edge(x, y):
                if bool(graph.get_edge_data(x, y)):
                    if y in graph.get_edge_data(x, y).keys():
                        mat[labels.index(x)][labels.index(y)] = symbol_map[
                            graph.get_edge_data(x, y)[y]
                        ]
                    else:
                        weight = graph.get_edge_data(x, y)[weight_label]
                        if weight is None:
                            weight = 1
                        mat[labels.index(x)][labels.index(y)] = weight
                else:
                    mat[labels.index(x)][labels.index(y)] = 1
    mat[np.isnan(mat)] = 0
    return mat


def graph_to_adjacency_file(graph: AnyGraph, output_file: Union[Path, str], labels):
    """
    A method to write the adjacency matrix of the graph to a file. If graph has
    weights, these are the values stored in the adjacency matrix.

    Args:
        graph: (Union[Graph, DiGraph] the graph to be saved
        output_file: (str) The full path where graph is to be saved
    """
    mat = graph_to_adjacency(graph, labels)
    with open(output_file, "w", encoding="utf-8") as f:
        f.write(",".join([f"{label}" for label in labels]))
        f.write("\n")
        for i, label in enumerate(labels):
            f.write(f"{label}")
            f.write(",")
            f.write(",".join([str(point) for point in mat[i]]))
            f.write("\n")

    return


def graph_to_dot_file(graph: AnyGraph, output_file: Union[Path, str]):
    """
    A method to write the graph in dot format to a file.

    Args:
        graph: (Union[Graph, DiGraph] the graph to be saved
        output_file: (str) The full path where graph is to be saved
    """
    nx.drawing.nx_pydot.write_dot(graph, output_file)


def select_device(force: Optional[str] = None) -> str:
    """
    Selects the device to be used for training. If force is not None, then
    the device is forced to be the one specified. If force is None, then
    the device is selected based on the availability of GPUs. If no GPUs are
    available, then the CPU is selected.

    Args:
        force (str): If not None, then the device is forced to be the one
            specified. If None, then the device is selected based on the
            availability of GPUs. If no GPUs are available, then the CPU is
            selected.

    Returns:
        (str) The device to be used for training.

    Raises:
        ValueError: If the forced device is not available or not a valid device.
    """
    if force is not None:
        if force in ['cuda', 'mps', 'cpu']:
            if force == 'cuda' and torch.cuda.is_available():
                device = force
            elif force == 'mps' and torch.backends.mps.is_available():
                device = force
            elif force == 'mps' and not torch.backends.mps.is_available():
                device = 'cpu'
            elif force == 'cpu':
                device = force
            else:
                raise ValueError(f"Invalid device: {force}")
        else:
            raise ValueError(f"Invalid device: {force}")
    else:
        if torch.cuda.is_available() and torch.backends.cuda.is_built():
            device = "cuda"
        elif torch.backends.mps.is_available() and torch.backends.mps.is_built():
            device = "mps"
        else:
            device = "cpu"
    return device


def graph_intersection(g1: nx.DiGraph, g2: nx.DiGraph) -> nx.DiGraph:
    """
    Returns the intersection of two graphs. The intersection is defined as the
    set of nodes and edges that are common to both graphs. The intersection is
    performed on the nodes and edges, not on the attributes of the nodes and
    edges.

    Args:
        g1 (networkx.DiGraph): The first graph.
        g2 (networkx.DiGraph): The second graph.

    Returns:
        (networkx.DiGraph) The intersection of the two graphs.
    """
    # Get the nodes from g1 and g2, as the union of both
    nodes = set(g1.nodes).intersection(set(g2.nodes))

    # Check if any of the graphs is empty
    if len(nodes) == 0:
        return nx.DiGraph()

    # For the nodes in the resulting DAG ("g"), get the data from the nodes in
    # g1 and g2, as the average value of both
    first_node = list(g1.nodes)[0]
    data_fields = g1.nodes[first_node].keys()
    nodes_data = {}
    for key in data_fields:
        # Check if the value for this key is numeric
        if isinstance(g1.nodes[first_node][key], (int, float, np.number)):
            for n in nodes:
                if n in g1.nodes and n in g2.nodes:
                    nodes_data[n] = {
                        key: (g1.nodes[n][key] + g2.nodes[n][key]) / 2}
                elif n in g1.nodes:
                    nodes_data[n] = {key: g1.nodes[n][key]}
                elif n in g2.nodes:
                    nodes_data[n] = {key: g2.nodes[n][key]}

    # Get the edges from g1 and g2, and take only those that match completely
    edges = g1.edges & g2.edges

    # Create a new graph with the nodes, nodes_data and edges
    g = nx.DiGraph()
    g.add_nodes_from(nodes)
    g.add_edges_from(edges)
    nx.set_node_attributes(g, nodes_data)

    return g


def graph_union(g1: nx.DiGraph, g2: nx.DiGraph) -> nx.DiGraph:
    """
    Returns the union of two graphs. The union is defined as the set of nodes and
    edges that are in both graphs. The union is performed on the nodes and edges,
    not on the attributes of the nodes and edges.
    """
    nodes = set(g1.nodes).union(set(g2.nodes))
    g = nx.DiGraph()
    g.add_nodes_from(nodes)
    g.add_edges_from(g1.edges)
    g.add_edges_from(g2.edges)

    # For the nodes in the resulting DAG ("g"), get the data from the nodes in
    # g1 and g2, as the average value of both
    first_node = list(g1.nodes)[0]
    data_fields = g1.nodes[first_node].keys()
    nodes_data = {}
    for key in data_fields:
        # Check if the value for this key is numeric
        if isinstance(g1.nodes[first_node][key], (int, float, np.number)):
            for n in nodes:
                if n in g1.nodes and n in g2.nodes:
                    nodes_data[n] = {
                        key: (g1.nodes[n][key] + g2.nodes[n][key]) / 2}
                elif n in g1.nodes:
                    nodes_data[n] = {key: g1.nodes[n][key]}
                elif n in g2.nodes:
                    nodes_data[n] = {key: g2.nodes[n][key]}
    nx.set_node_attributes(g, nodes_data)

    return g


def digraph_from_connected_features(
        X,
        feature_names,
        models,
        connections,
        root_causes,
        prior: Optional[List[List[str]]] = None,
        reciprocity=True,
        anm_iterations=10,
        max_anm_samples=400,
        verbose=False):
    """
    Builds a directed graph from a set of features and their connections. The
    connections are determined by the SHAP values of the features. The
    orientation of the edges is determined by the causal direction of the
    features. The orientation is determined by the method of edge orientation
    defined in the independence module.

    Parameters:
    -----------
    X (numpy.ndarray): The input data.
    feature_names (list): The list of feature names.
    models (obj): The model used to explain the data.
    connections (dict): The dictionary of connections between features. The
        dictionary is of the form {'feature1': ['feature2', 'feature3'], ...}
    root_causes (list): The list of root causes. The root causes are the
        features that are not caused by any other feature.
    reciprocity (bool): If True, then the edges are oriented only if the
        direction is reciprocal. If False, then the edges are oriented
        regardless of the reciprocity.
    anm_iterations=10 (int): The number of iterations to be used by the independence
        module to determine the orientation of the edges.
    max_anm_samples=400 (int): The maximum number of samples to be used by the
        independence module to determine the orientation of the edges. If the
        number of samples is larger than the number of samples in the data, then
        the number of samples is set to the number of samples in the data.
    verbose (bool): If True, then the function prints information about the
        orientation of the edges.

    Returns:
    --------
    networkx.DiGraph: The directed graph with the nodes and edges determined
        by the connections and the orientation determined by the independence
        module.

    """
    if verbose:
        print("-----\ndigraph_from_connected_features()")
    unoriented_graph = nx.Graph()
    for target in feature_names:
        for peer in connections[target]:
            # Add edges ONLY between nodes where SHAP recognizes both directions
            if reciprocity:
                if target in connections[peer]:
                    unoriented_graph.add_edge(target, peer)
            else:
                unoriented_graph.add_edge(target, peer)

    dag = nx.DiGraph()
    # Add regression mean score to each node
    for i, feature in enumerate(feature_names):
        dag.add_node(feature)
        dag.nodes[feature]['regr_score'] = models.scoring[i]

    # Determine edge orientation for each edge
    if X.shape[0] > max_anm_samples:
        X = X.sample(max_anm_samples)
        if verbose:
            print(f"Reduced number of samples to {max_anm_samples}")
    for u, v in unoriented_graph.edges():
        # Set orientation to 0 ~ unknown
        orientation = 0

        # Check if we can set it from prior knowledge
        if prior:
            orientation = correct_edge_from_prior(dag, u, v, prior, verbose)

        if orientation == 0:
            print(f"Checking edge {u} -> {v}...") if verbose else None
            # Import locally to avoid requiring heavy deps (hyppo/numba) unless needed
            from ..independence.edge_orientation import get_edge_orientation
            orientation = get_edge_orientation(
                X, u, v, iters=anm_iterations, method="gpr", verbose=verbose)
            if orientation == +1:
                print(f"  Edge {u} -> {v} added from ANM") if verbose else None
                dag.add_edge(u, v)
            elif orientation == -1:
                print(f"  Edge {v} -> {u} added from ANM") if verbose else None
                dag.add_edge(v, u)
            else:
                pass

    # Apply a simple fix: if quality of regression for a feature is poor, then
    # that feature can be considered a root or parent node. Therefore, we cannot
    # have an edge pointing to that node.
    if root_causes:
        changes = []
        for parent_node in root_causes:
            print(f"Checking root cause {parent_node}...") if verbose else None
            for cause, effect in dag.edges():
                if effect == parent_node:
                    print(
                        f"Reverting edge {cause} -> {effect}") if verbose else None
                    changes.append((cause, effect))
        for cause, effect in changes:
            dag.remove_edge(cause, effect)
            dag.add_edge(effect, cause)

    return dag


def correct_edge_from_prior(dag, u, v, prior, verbose):
    """
    Correct an edge in the graph according to the prior knowledge.
    This function corrects the orientation of an edge in a directed acyclic graph
    (DAG) based on prior knowledge. The prior knowledge is a list of lists of node
    names, ordered according to a temporal structure. The assumption is that nodes
    from the first layer of temporal structure cannot receive any incoming edge.

    The function checks the relative positions of the two nodes in the prior
    knowledge and:
    - If both nodes are in the top list, it removes the edge, based on the assumption.
    - If one node is before the other, it adds a new edge in the correct direction.
    - If the nodes are not in a clear order, it leaves the edge unchanged.
    - If the edge reflects a backward connection between nodes from different
      temporal layers, it removes the edge.

    The orientation of the edge, -1 if the edge is reversed, +1 if the edge
        is removed or kept, and 0 if the edge orientation is not clear.

    Parameters
    ----------
    dag : nx.DiGraph
        The graph to be corrected.
    u : str
        The first node of the edge.
    v : str
        The second node of the edge.
    prior : List[List[str]]
        The prior knowledge, a list of lists of node names, ordered according to
        a temporal structure.
    verbose : bool
        If True, print debug messages.

    Returns
    -------
    orientation : int
        The orientation of the edge, -1 if the edge is reversed, +1 if the edge
        is removed or kept, and 0 if the edge orientation is not clear.
    """
    # Check either 'u' or 'v' are NOT in the prior list. Remember that 'prior' is a
    # list of lists of node names, ordered according to a temporal structure.
    if not any([u in p for p in prior]) or not any([v in p for p in prior]):
        return 0

    print(f"Checking edge {u} -> {v}...") if verbose else None
    idx_u = [i for i, l in enumerate(prior) if u in l][0]
    idx_v = [i for i, l in enumerate(prior) if v in l][0]
    both_in_top_list = idx_u == 0 and idx_v == 0
    u_is_before_v = idx_u - idx_v < 0
    v_is_before_u = idx_v - idx_u < 0
    if both_in_top_list:
        print(
            f"Edge {u} -x- {v} removed: both top list") if verbose else None
        dag.remove_edge(u, v) if dag.has_edge(u, v) else None # Experimental XXX Beware of this line!
        return +1
    elif u_is_before_v:
        print(
            f"Edge {u} -> {v} added: {u} before {v}") if verbose else None
        dag.add_edge(u, v) if not dag.has_edge(v, u) else None
        return +1
    elif v_is_before_u:
        print(
            f"Edge {v} -> {u} added: {v} before {u}") if verbose else None
        dag.remove_edge(u, v) if dag.has_edge(u, v) else None
        return -1
    else:
        return 0


def valid_candidates_from_prior(feature_names, effect, prior):
    """
    This method returns the valid candidates for a given effect, based on the
    prior information. The prior information is a list of lists, where each list
    contains the features that are known to be in the same level of the hierarchy.

    Parameters:
    -----------
    feature_names: List[str]
        The list of feature names.
    effect: str
        The effect for which to find the valid candidates.
    prior: List[List[str]]
        The prior information about the hierarchy of the features.

    Returns:
    --------
    List[str]
        The valid candidates for the given effect.
    """
    if prior is None or prior == []:
        return [c for c in feature_names if c != effect]

    # identify in what list is the effect, from the list of lists defined in prior
    idx = [i for i, sublist in enumerate(prior) if effect in sublist]
    if not idx:
        raise ValueError(f"Effect '{effect}' not found in prior")

    # candidates are elements in the list ' idx' and all the lists before it
    candidates = [item for sublist in prior[:idx[0] + 1]
                  for item in sublist]

    # return the candidates, excluding the effect itself
    return [c for c in candidates if c != effect]


def break_cycles_using_prior(
        original_dag: nx.DiGraph,
        prior: List[List[str]],
        verbose: bool = False) -> nx.DiGraph:
    """
    This method remove potential edges in the DAG by removing edges that are
    incompatible with the temporal relationship defined in the prior knowledge.
    Any edge pointing backwards, according to the order established in the prior
    knowledge, is removed.

    Parameters:
    -----------
    original_dag (nx.DiGraph): The original directed acyclic graph.
    prior (List[List[str]]): A list of lists containing the prior knowledge
        about the edges in the DAG. The lists define a hierarchy of edges that
        represent a temporal relation in cause and effect. If a node is in the first
        list, then it is a root cause. If a node is in the second list, then it is
        caused by the nodes in the first list or the second list, and so on.
    verbose (bool): If True, then the method prints information about the
        edges that are removed.

    Returns:
    --------
    nx.DiGraph: The directed acyclic graph with the edges removed according to
        the prior knowledge.
    """
    if prior is None or prior == []:
        return original_dag

    new_dag = original_dag.copy()

    if verbose:
        print("Prior knowledge:", prior)
    cycles = list(nx.simple_cycles(new_dag))
    while len(cycles) > 0:
        cycle = cycles.pop(0)
        if verbose:
            print("Checking cycle", cycle)
        for i, node in enumerate(cycle):
            if node == cycle[-1]:
                neighbour = cycle[0]
            else:
                neighbour = cycle[i+1]
            # Check if the edge between the two nodes is in the prior knowledge
            if [node, neighbour] not in prior:
                if verbose:
                    print(
                        f"↳ Checking '{node} -> {neighbour}' in the prior knowledge")
                idx_node = [i for i, l in enumerate(prior) if node in l][0]
                idx_neighbour = [i for i, l in enumerate(
                    prior) if neighbour in l][0]
                if idx_neighbour - idx_node < 0:
                    if verbose:
                        print(
                            f"  ↳ Breaking cycle, removing edge {node} -> {neighbour}")
                        print("** Recomputing cycles **")
                    new_dag.remove_edge(node, neighbour)
                    cycles = list(nx.simple_cycles(new_dag))
                    break
    return new_dag


def potential_misoriented_edges(
        loop: List[str],
        discrepancies: Dict,
        verbose: bool = False) -> List[Tuple[str, str, float]]:
    """
    Find potential misoriented edges in a loop based on discrepancies. The loop
    is a list of nodes that form a loop in a DAG. The discrepancies is a dictionary
    containing the discrepancies between nodes. The function returns a list of
    potential misoriented edges, sorted by the difference in goodness-of-fit
    scores.

    Args:
        loop (List[str]): The list of nodes in the loop.
        discrepancies (Dict): A dictionary containing discrepancies between
            nodes. It is typically a positive float number between 0 and 1.
        verbose (bool, optional): Whether to print verbose output.
            Defaults to False.

    Returns:
        List[Tuple[str, str, float]]: A list of potential misoriented edges,
            sorted by the difference in goodness-of-fit scores.

    """
    potential_misoriented_edges = []
    for i, node in enumerate(loop):
        if i == len(loop) - 1:
            neighbor = loop[0]
        else:
            neighbor = loop[i + 1]
        if neighbor not in discrepancies or node not in discrepancies:
            continue
        if node not in discrepancies[neighbor] or neighbor not in discrepancies[node]:
            continue
        fwd_gof = 1. - discrepancies[neighbor][node].shap_gof
        bwd_gof = 1. - discrepancies[node][neighbor].shap_gof
        orientation = +1 if fwd_gof < bwd_gof else -1
        if orientation == -1:
            potential_misoriented_edges.append(
                (node, neighbor, fwd_gof - bwd_gof))
        if verbose:
            direction = "-->" if fwd_gof < bwd_gof else "<--"
            print(
                f"Edge: {node}{direction}{neighbor} FWD:{fwd_gof:.3f} "
                f"BWD:{bwd_gof:.3f}")

    return sorted(potential_misoriented_edges, key=lambda x: x[2], reverse=True)


def break_cycles_if_present(
        dag: nx.DiGraph,
        discrepancies: Dict,
        prior: Optional[List[List[str]]] = None,
        verbose: bool = False) -> nx.DiGraph:
    """
    Breaks cycles in a directed acyclic graph (DAG) by removing the edge with
    the lowest goodness of fit (R2). If there are multiple cycles, they are
    all traversed and fixed. If prior is set, then the cycles are broken
    using the prior knowledge.

    Parameters:
    - dag (nx.DiGraph): the DAG to break cycles in.
    - knowledge (pd.DataFrame): a DataFrame containing the permutation importances
        for each edge in the DAG.
    - prior (List[List[str]]): a list of lists containing the prior knowledge
        about the edges in the DAG. The lists define a hierarchy of edges that
        represent a temporal relation in cause and effect. If a node is in the first
        list, then it is a root cause. If a node is in the second list, then it is
        caused by the nodes in the first list or the second list, and so on.

    Returns:
    - dag (nx.DiGraph): the DAG with cycles broken.
    """
    if verbose:
        print("-----\nbreak_cycles_if_present()")

    # If prior is set, then break cycles using the prior knowledge
    if prior:
        new_dag = break_cycles_using_prior(dag, prior, verbose)
    else:
        new_dag = dag.copy()

    # Find all cycles in the DAG
    cycles = list(nx.simple_cycles(new_dag))

    # This might be important, to remove first double edges
    cycles.sort(key=len)
    if len(cycles) == 0:
        if verbose:
            print("No cycles found")
        return new_dag

    # Traverse all cycles, fixing them
    cycles_info = []
    while len(cycles) > 0:
        cycle = cycles.pop(0)
        # For every pair of consecutive nodes in the cycle, store its SHAP
        # discrepancy
        cycle_edges = {}
        for node in cycle:
            if node == cycle[-1]:
                neighbour = cycle[0]
            else:
                neighbour = cycle[cycle.index(node)+1]
            cycle_edges[(node, neighbour)
                        ] = discrepancies[neighbour][node].shap_gof
        cycles_info.append((cycle, cycle_edges))

        # Find the edge with the lowest SHAP discrepancy
        min_gof = min(cycle_edges.values())
        min_edge = [edge for edge, gof in cycle_edges.items() if gof ==
                    min_gof][0]

        # Check if any of the edges in the loop is potentially misoriented.
        # If so, change the orientation of the edge with the lowest SHAP
        # discrepancy
        potential_misoriented = potential_misoriented_edges(
            cycle, discrepancies, verbose)
        if len(potential_misoriented) > 0:
            if verbose:
                print("Potential misoriented edges in the cycle:")
                for edge in potential_misoriented:
                    print(f"  {edge[0]} --> {edge[1]} ({edge[2]:.3f})")
            # Check if changing the orientation of the edge would break
            # the cycle
            test_dag = new_dag.copy()

            # Check if the edge is yet in the test_dag
            if min_edge in test_dag.edges:
                test_dag.remove_edge(*min_edge)

            test_dag.add_edge(*potential_misoriented[0][1::-1])
            test_cycles = list(nx.simple_cycles(test_dag))
            if len(test_cycles) == len(cycles):
                if verbose:
                    print(
                        f"Breaking cycle {cycle} by changing orientation of "
                        f"edge {min_edge}")
                new_dag.remove_edge(*min_edge)
                new_dag.add_edge(*potential_misoriented[0][1::-1])
                continue

        # Remove the edge with the lowest SHAP discrepancy, checking that
        # the edge still exists
        if min_edge in new_dag.edges:
            if verbose:
                print(f"Breaking cycle {cycle} by removing edge {min_edge}")
            new_dag.remove_edge(*min_edge)

        # Recompute whether there're cycles in the DAG after this change
        cycles = list(nx.simple_cycles(new_dag))
        cycles.sort(key=len)

    return new_dag


def stringfy_object(object_: object) -> str:
    """
    Convert an object into a string representation, including its attributes.

    Args:
        object_ (object): The object to be converted.

    Returns:
        str: A string representation of the object and its attributes.
    """
    forbidden_attrs = [
        'fit', 'predict', 'fit_predict', 'score', 'get_params', 'set_params']
    ret = "Object attributes\n"
    ret += f"{'-'*80}\n"
    for attr in dir(object_):
        if attr.startswith('_') or \
            attr in forbidden_attrs or \
                isinstance(getattr(object_, attr), types.MethodType):
            continue
        if isinstance(getattr(object_, attr), (pd.DataFrame, np.ndarray)):
            ret += f"{attr:25} {type(getattr(object_, attr)).__name__} "
            ret += f"{getattr(object_, attr).shape}\n"
            continue
        if isinstance(getattr(object_, attr), nx.DiGraph):
            n_nodes = getattr(object_, attr).number_of_nodes()
            n_edges = getattr(object_, attr).number_of_edges()
            ret += f"{attr:25} {n_nodes} nodes, {n_edges} edges\n"
            continue
        if isinstance(getattr(object_, attr), dict):
            keys_list = [
                f"{k}:{type(getattr(object_, attr)[k])}"
                for k in getattr(object_, attr).keys()
            ]
            ret += f"{attr:25} dict {str(keys_list)[:60]}...\n"
            continue
        if isinstance(getattr(object_, attr), list):
            list_ = '[' + ', '.join(str(e)
                                    for e in getattr(object_, attr)) + ']'
            ret += f"{attr:25} list {list_[:60]}...\n"
            continue
        if isinstance(getattr(object_, attr), (int, float, str, bool)):
            ret += f"{attr:25} {getattr(object_, attr)}\n"
            continue
        ret += f"{attr:25} <{getattr(object_, attr).__class__.__name__}>\n"

    return ret


def get_feature_names(X: Union[pd.DataFrame, np.ndarray, list]) -> List[str]:
    """
    Get the feature names from the input data. The feature names can be
    extracted from a pandas DataFrame, a numpy array or a list.

    Parameters:
    -----------
    X : Union[pd.DataFrame, np.ndarray, list]
        The input data.

    Returns:
    --------
    List[str]
        The list of feature names.
    """
    assert isinstance(X, (pd.DataFrame, np.ndarray, list)), \
        "X must be a pandas DataFrame, a numpy array or a list"

    if isinstance(X, pd.DataFrame):
        feature_names = list(X.columns)
    elif isinstance(X, np.ndarray):
        feature_names = [f"X{i}" for i in range(X.shape[1])]
    else:
        feature_names = [f"X{i}" for i in range(len(X[0]))]

    return feature_names


def _classify_variable(arr):
    """
    This method inspect the contents of the array and returns the type of the
    variable. If the variables are NOT numbers, then it returns multiclass. If
    the variable is a number, then it returns binary if it has only two unique
    values. If the number of unique values is greater than two, then it returns
    continuous.

    Args:
        arr (Union[pd.Series, np.ndarray]): The array to be inspected.

    Returns:
        str: The type of the variable. Can be binary, multiclass or continuous.
    """
    # Convert numpy array to pandas Series if needed
    if isinstance(arr, np.ndarray):
        unique_values = np.unique(arr)
    else:
        unique_values = arr.unique()

    # Handle empty series
    if len(unique_values) == 0:
        return "continuous"

    # Check if the variable is not a number
    if unique_values.dtype.kind not in "biufc":
        return "multiclass"
    # Check if the variable has only two unique values
    elif len(unique_values) == 2:
        return "binary"
    else:
        return "continuous"


def get_feature_types(X) -> Dict[str, str]:
    """
    Get the feature types from the input data. The feature types can be
    binary, multiclass or continuous. The classification is done based on the
    number of unique values in the array. If the array has only two unique
    values, then the variable is classified as binary. If the unique values
    are integers, then the variable is classified as multiclass. Otherwise,
    the variable is classified as continuous.

    Parameters:
    -----------
    X : Union[pd.DataFrame, np.ndarray, list]
        The input data.

    Returns:
    --------
    Dict[str, str]
        A dictionary with the feature names as keys and the feature types as
        values.

    Example:
    --------
    >>> X = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    >>> get_feature_types(X)
    {'X0': 'multiclass', 'X1': 'multiclass', 'X2': 'multiclass'}
    """
    feature_names = get_feature_names(X)
    feature_types = {}
    if isinstance(X, np.ndarray):
        for i, feature in enumerate(X.T):
            feature_types[feature_names[i]] = _classify_variable(feature)
    elif isinstance(X, pd.DataFrame):
        for feature in X.columns:
            feature_types[feature] = _classify_variable(X[feature].values)
    else:
        for i, feature in enumerate(X[0]):
            feature_types[feature_names[i]] = _classify_variable(feature)

    return feature_types


def cast_categoricals_to_int(X: pd.DataFrame) -> pd.DataFrame:
    """
    Cast all categorical features in the dataframe to integer values.

    Parameters
    ----------
    X: pd.DataFrame
        The dataframe to cast.

    Returns
    -------
    pd.DataFrame
        The dataframe with all categorical features cast to integer values.
    """
    X = X.copy()
    for feature in X.columns:
        feature_dtype = _classify_variable(X[feature])
        if feature_dtype == "multiclass" or feature_dtype == "binary":
            # Convert all the values to integers. For example, 'a' -> 0, 'b' ->
            # 1, 'c' -> 2, respecting the order of the unique values.abs
            X[feature] = X[feature].astype("category")
            X[feature] = X[feature].cat.codes
            X[feature] = X[feature].astype("int16")

    return X


def find_crossing_point(
        f1: List[float],  # Values of the first curve (precision, for example)
        f2: List[float],  # Values of the second curve (recall, for example)
        # X values (thresholds, for example) corresponding
        x_values: List[float]
    # to the curve evaluations
) -> Optional[float]:
    """
    This function finds the exact crossing point between two curves defined by
    f1 and f2 over x_values. It interpolates between points to provide a more
    accurate crossing point.

    Parameters:
    -----------
    f1: List[float]
        Values of the first curve (precision, for example).
    f2: List[float]
        Values of the second curve (recall, for example).
    x_values: List[float]
        X values (thresholds, for example) corresponding to the curve evaluations.

    Returns:
    --------
    Optional[float]
        The x value where the first crossing occurs between the two curves.
        If no crossing is found, returns None.
    """
    def interpolate_crossing(
            f1_prev: float,  # Value of f1 at the previous x value
            f2_prev: float,  # Value of f2 at the previous x value
            x_prev: float,  # Previous x value
            x_next: float,  # Next x value
            f1_next: float,  # Value of f1 at the next x value
            f2_next: float  # Value of f2 at the next x value
    ) -> float:
        # Linear interpolation formula to find the exact crossing point between
        # two points
        return x_prev + (x_next - x_prev) * abs(f1_prev - f2_prev) / \
            (abs(f1_prev - f1_next) + abs(f2_prev - f2_next))

    # Loop through the x values and find where the first crossing occurs
    for i in range(1, len(x_values)):
        # Check for a crossing between consecutive points
        if (f1[i-1] < f2[i-1] and f1[i] > f2[i]) or \
                (f1[i-1] > f2[i-1] and f1[i] < f2[i]):
            # Interpolate to find the exact crossing point
            return interpolate_crossing(
                f1[i-1], f2[i-1], x_values[i-1], x_values[i], f1[i], f2[i])

    return None  # No crossing found


def format_time(seconds: float) -> (float, str):  # type: ignore
    """
    Convert the time in seconds to a more convenient time unit.

    Parameters
    ----------
    seconds : float
        The time in seconds.

    Returns
    -------
    (float, str)
        The time in the most appropriate unit and the unit string.

    Examples
    --------
    >>> format_time(1.0)
    (1.0, 's')
    >>> format_time(60.0)
    (1.0, 'm')
    >>> format_time(3600.0)
    (1.0, 'h')
    >>> format_time(86400.0)
    (1.0, 'd')
    >>> format_time(604800.0)
    (1.0, 'w')
    >>> format_time(2592000.0)
    (1.0, 'm')
    >>> format_time(31536000.0)
    (1.0, 'y')
    >>> format_time(315360000.0)
    (1.0, 'a')
    """
    if seconds < 0:
        raise ValueError("time cannot be negative")
    if seconds < 60.0:
        return seconds, 's'
    elif seconds < 3600.0:
        return seconds / 60.0, 'm'
    elif seconds < 86400.0:
        return seconds / 3600.0, 'h'
    elif seconds < 604800.0:
        return seconds / 86400.0, 'd'
    elif seconds < 2592000.0:
        return seconds / 604800.0, 'w'
    elif seconds < 31536000.0:
        return seconds / 2592000.0, 'm'
    elif seconds < 315360000.0:
        return seconds / 31536000.0, 'y'
    else:
        return seconds / 315360000.0, 'a'


def combine_dags(
    dag1: nx.DiGraph,
    dag2: nx.DiGraph,
    discrepancies: Dict,
    prior: Optional[List[List[str]]] = None
) -> Tuple[nx.DiGraph, nx.DiGraph, nx.DiGraph, nx.DiGraph]:
    """
    Combine two directed acyclic graphs (DAGs) into a single DAG.

    Parameters
    ----------
    dag1 : nx.DiGraph
        The first DAG.
    dag2 : nx.DiGraph
        The second DAG.
    discrepancies : Dict
        A Dictionary containing the permutation importances for each edge in the DAGs.
    prior : Optional[List[List[str]]], optional
        A list of lists containing the prior knowledge about the edges in the DAGs.
        The lists define a hierarchy of edges that represent a temporal relation in
        cause and effect. If a node is in the first list, then it is a root cause.
        If a node is in the second list, then it is caused by the nodes in the
        first list or the second list, and so on.

    Returns
    -------
    Tuple[nx.DiGraph, nx.DiGraph, nx.DiGraph, nx.DiGraph]
        A tuple containing four graphs: the union of the two DAGs, the
        intersection of the two DAGs, the union of the two DAGs after removing
        cycles, and the intersection of the two DAGs after removing cycles.
    """
    # Compute the union of the two DAGs
    union = graph_union(dag1, dag2)

    # Compute the intersection of the two DAGs
    inter = graph_intersection(dag1, dag2)

    # Remove cycles from the union and intersection
    union_cycles_removed = break_cycles_if_present(
        union, discrepancies, prior)
    inter_cycles_removed = break_cycles_if_present(
        inter, discrepancies, prior)

    # XXX: Experiment -> Use prior to correct edges in the union
    if prior:
        edges_list = list(union_cycles_removed.edges)
        for u, v in edges_list:
            correct_edge_from_prior(union_cycles_removed, u, v, prior, verbose=False)

    return union, inter, union_cycles_removed, inter_cycles_removed


def list_files(input_pattern: str, where: str) -> list:
    """
    List all the files in the input path matching the input pattern

    Parameters:
    -----------
    input_pattern : str
        The pattern to match the files
    where : str
        The path to use to look for the files matching the pattern
    """
    input_files = glob.glob(os.path.join(
        where, input_pattern))
    input_files = sorted([os.path.basename(os.path.splitext(f)[0])
                         for f in input_files])

    assert len(input_files) > 0, \
        f"No files found in {where} matching <{input_pattern}>"

    return sorted(list(set(input_files)))

def read_json_file(file_path: str):
    """
    Read a JSON file and return its content as a dictionary."
    """
    try:
        with open(file_path, 'r', encoding='utf-8') as file:
            data = json.load(file)
            # Get 'prior' key, default to empty list
            prior = data.get("prior", [])
            return prior
    except FileNotFoundError:
        print(f"Error: File '{file_path}' not found.")
        return []
    except json.JSONDecodeError:
        print("Error: Invalid JSON format.")
        return []



import pytest
import numpy as np

from causalexplain.independence.hsic import HSIC
from hyppo.independence import Hsic


@pytest.fixture
def vectors():
    X = np.array([
        0.136486655, 0.108931511, 0.022105488, 0.349472863, 0.148249433,
        -0.321564702, 0.320629400, -0.282624440, 0.263522936, -0.316252276])
    Y = np.array([
        -0.1669332713, 0.4886635816, 0.5315530519, 0.1330376544, -0.0632027887,
        0.1640341743, -0.1833757726, 0.3754725901, -0.0722728821, -0.0395241960])
    return X, Y


# This is the test from causalexplain
def test_HSIC(vectors):
    X, Y = vectors

    np.random.seed(1234)
    h = HSIC()
    result = h.fit(X, Y)
    assert result.hsic == pytest.approx(0.0)
    assert result.stat == pytest.approx(0.0)
    assert result.p_value == pytest.approx(0.33058547413699446)
    assert result.independence


# This is the test from hyppo. p-value matches the one from causalexplain
def test_HSIC_own(vectors):
    X, Y = vectors

    np.random.seed(1234)
    stat, pval = Hsic().test(X, Y)
    assert stat == pytest.approx(0.18, abs=0.01)
    assert pval == pytest.approx(0.30, abs=0.01)



import numpy as np
import pandas as pd
import pytest

from causalexplain.estimators.cam.cam import CAM


def test_cam_requires_fit_before_predict():
    model = CAM(name="test")
    with pytest.raises(ValueError):
        model.predict()


def test_cam_fit_and_predict_flow(monkeypatch):
    def fake_compute_score_mat(*args, **kwargs):
        return {
            "scoreMat": np.array([[-np.inf, 0.5], [0.1, -np.inf]]),
            "rowParents": np.array([[0], [1]]),
            "scoreEmptyNodes": np.zeros(2),
        }

    monkeypatch.setattr("causalexplain.estimators.cam.cam.computeScoreMat", fake_compute_score_mat)
    monkeypatch.setattr(
        "causalexplain.estimators.cam.cam.updateScoreMat",
        lambda score_mat, *args, **kwargs: np.full_like(score_mat, -np.inf),
    )
    monkeypatch.setattr(
        "causalexplain.estimators.cam.cam.pruning",
        lambda X, G, prune_method=None, prune_method_pars=None, verbose=False: G,
    )
    fake_graph = object()
    monkeypatch.setattr(
        "causalexplain.estimators.cam.cam.utils.graph_from_adjacency",
        lambda adj, names: fake_graph,
    )
    monkeypatch.setattr(
        "causalexplain.estimators.cam.cam.evaluate_graph",
        lambda ref_graph, dag, feature_names: {"dummy": 1},
    )

    data = pd.DataFrame([[0.0, 1.0], [1.0, 0.0]], columns=["a", "b"])
    model = CAM(name="test_cam", pruning=True, verbose=False)

    fitted = model.fit(data)

    assert fitted.is_fitted_
    assert fitted.adj[0, 1] == 1
    assert fitted.adj[1, 0] == 0

    dag = fitted.predict()
    assert dag is fake_graph



import numpy as np
import pytest

from causalexplain.estimators.cam import updateScoreMat as update_score_mat_module


def test_update_score_mat_recomputes_allowed_entries(monkeypatch):
    calls = []

    def fake_parallel(row_parents, sel_mat, score_name, X, verbose, node2, pars_score, interv_mat, interv_data, i):
        calls.append((row_parents.copy(), node2, i))
        return 2.0

    monkeypatch.setattr(update_score_mat_module, "computeScoreMatParallel", fake_parallel)

    score_mat = np.zeros((3, 3))
    X = np.ones((4, 3))
    adj = np.zeros((3, 3))
    adj[0, 1] = 1

    updated = update_score_mat_module.updateScoreMat(
        score_mat=score_mat,
        X=X,
        score_name="test",
        i=0,
        j=1,
        score_nodes=np.zeros(3),
        adj=adj,
        verbose=False,
        num_cores=1,
        max_num_parents=2,
        pars_score={},
        interv_mat=None,
        interv_data=False,
    )

    assert updated[2, 1] == 2.0
    assert calls[0][1:] == (1, 2)


def test_update_score_mat_respects_parent_limit(monkeypatch):
    monkeypatch.setattr(
        update_score_mat_module,
        "computeScoreMatParallel",
        lambda *args, **kwargs: pytest.fail("Should not be called when parent limit reached"),
    )

    score_mat = np.zeros((3, 3))
    X = np.ones((4, 3))
    adj = np.zeros((3, 3))
    adj[0, 1] = 1

    updated = update_score_mat_module.updateScoreMat(
        score_mat=score_mat,
        X=X,
        score_name="test",
        i=0,
        j=1,
        score_nodes=np.zeros(3),
        adj=adj,
        verbose=False,
        num_cores=1,
        max_num_parents=0,
        pars_score={},
        interv_mat=None,
        interv_data=False,
    )

    assert updated[2, 1] == -np.inf



import numpy as np
import pytest

import causalexplain.independence.feature_selection as feature_selection
from causalexplain.independence.feature_selection import (
    find_cluster_change_point,
    select_features,
)


def test_select_features_requires_threshold_when_exhaustive():
    with pytest.raises(ValueError):
        select_features(np.array([1.0, 2.0]), ["a", "b"], exhaustive=True)


def test_select_features_exits_when_impact_below_minimum():
    values = np.array([1e-8, 2e-8, 3e-8])
    assert select_features(values, ["a", "b", "c"], min_impact=1e-6) == []


def test_select_features_handles_matrix_and_sorts_by_mean():
    values = np.array([[0.0, 0.1, 0.9], [0.0, 0.1, 0.9]])
    assert select_features(values, ["a", "b", "c"]) == ["c"]


def test_select_features_returns_shaps_when_requested():
    values = np.array([0.0, 0.0, 1.0, 2.0])
    selected, shaps = select_features(
        values,
        ["w", "x", "y", "z"],
        return_shaps=True,
        exhaustive=True,
        threshold=0.1,
    )
    assert selected == ["z", "y"]
    assert shaps == [2.0, 1.0]


def test_select_features_non_exhaustive_orders_by_cluster_change():
    values = np.array([0.1, 0.4, 0.2])
    assert select_features(values, ["a", "b", "c"]) == ["b"]


def test_select_features_exhaustive_iterates_until_threshold(monkeypatch):
    calls = []

    def fake_change_point(current_values, verbose=False):
        # Peel off one more element each time to exercise the exhaustive branch.
        calls.append(list(current_values))
        return len(current_values) - 1

    monkeypatch.setattr(feature_selection, "find_cluster_change_point", fake_change_point)
    values = np.array([0.1, 0.2, 0.3])
    result = select_features(values, ["a", "b", "c"], exhaustive=True, threshold=0.05)

    assert result == ["c", "b", "a"]
    assert calls == [[0.1, 0.2, 0.3], [0.1, 0.2], [0.1]]


def test_select_features_stops_after_max_iterations(monkeypatch):
    def non_progressing_change_point(current_values, verbose=False):
        # Returning len(current_values) means the slice never shrinks.
        return len(current_values)

    monkeypatch.setattr(
        feature_selection, "find_cluster_change_point", non_progressing_change_point
    )
    values = np.array([0.1, 0.2])
    assert select_features(values, ["a", "b"], exhaustive=True, threshold=0.05) == []


def test_select_features_emits_useful_verbose_output(capsys):
    values = np.array([0.1, 0.3, 1.0])
    select_features(values, ["a", "b", "c"], threshold=0.05, verbose=True)

    output = capsys.readouterr().out
    assert "Feature order" in output
    assert "Selected_features" in output


def test_select_features_verbose_matrix_output_lists_sums(capsys):
    values = np.array([[0.0, 0.2], [0.1, 0.3]])
    select_features(values, ["a", "b"], verbose=True)
    output = capsys.readouterr().out
    assert "Sum values" in output
    assert "(a:" in output


def test_find_cluster_change_point_handles_identical_values_and_logs(capsys):
    assert find_cluster_change_point([0, 0, 0], verbose=True) is None
    captured = capsys.readouterr()
    assert "** No clusters generated" in captured.out


def test_find_cluster_change_point_returns_none_for_short_series():
    assert find_cluster_change_point([]) is None
    assert find_cluster_change_point([1]) is None


def test_find_cluster_change_point_detects_gap_between_clusters():
    values = [0.1, 0.2, 0.8, 0.9]
    assert find_cluster_change_point(values) == 2


def test_find_cluster_change_point_reports_cluster_estimates_verbose(capsys):
    values = [0.1, 0.2, 0.8, 0.9]
    assert find_cluster_change_point(values, verbose=True) == 2
    output = capsys.readouterr().out
    assert "Est.clusters/noise" in output


def test_module_main_runs_and_logs(capsys):
    feature_selection.main()
    output = capsys.readouterr().out
    assert "Feature order" in output
    assert "threshold" in output


def test_embedded_test_function_is_intentionally_failing():
    with pytest.raises(AssertionError):
        feature_selection.test()



import types
import networkx as nx
import numpy as np
import pandas as pd
import pytest

from causalexplain.explainability import perm_importance as pimod


class DummyProgBar:
    def start_subtask(self, *args, **kwargs):
        return types.SimpleNamespace(update_subtask=lambda *a, **k: None, remove=lambda *a, **k: None)


class DummyEstimator:
    def __init__(self, name):
        self.name = name


class DummyModels:
    def __init__(self):
        self.regressor = {"a": DummyEstimator("a"), "b": DummyEstimator("b")}


def _patch_progbar(monkeypatch):
    monkeypatch.setattr(pimod, "ProgBar", lambda *a, **k: DummyProgBar())


def test_fit_sklearn_inserts_zero_for_correlated(monkeypatch):
    _patch_progbar(monkeypatch)
    monkeypatch.setattr(
        pimod.Hierarchies, "compute_correlation_matrix", staticmethod(lambda X: pd.DataFrame(np.eye(X.shape[1]), columns=X.columns, index=X.columns)),
    )
    monkeypatch.setattr(
        pimod.Hierarchies, "compute_correlated_features", staticmethod(lambda corr, th, names, verbose=False: {"a": ["b"], "b": ["a"]}),
    )
    monkeypatch.setattr(
        pimod,
        "permutation_importance",
        lambda estimator, X, y, n_repeats, random_state: {"importances_mean": np.array([]), "importances_std": np.array([])},
    )

    models = DummyModels()
    df = pd.DataFrame({"a": [1.0, 2.0, 3.0], "b": [0.5, 0.4, 0.3]})
    pi = pimod.PermutationImportance(models, correlation_th=0.8, mean_pi_percentile=0.5, prog_bar=False, verbose=False)
    pi.fit(df)

    assert pi.is_fitted_
    assert pi.pi["a"]["importances_mean"][0] == 0.0  # zero inserted for correlated 'b'
    assert pi.pi["b"]["importances_mean"][0] == 0.0  # zero inserted for correlated 'a'
    assert pi.mean_pi_threshold >= 0.0


def test_predict_sklearn_builds_graph(monkeypatch):
    _patch_progbar(monkeypatch)

    monkeypatch.setattr(
        pimod,
        "permutation_importance",
        lambda estimator, X, y, n_repeats, random_state: {"importances_mean": np.array([0.2]), "importances_std": np.array([0.05])},
    )
    monkeypatch.setattr(pimod.Hierarchies, "compute_correlation_matrix", staticmethod(lambda X: pd.DataFrame(np.eye(X.shape[1]), columns=X.columns, index=X.columns)))
    monkeypatch.setattr(pimod.Hierarchies, "compute_correlated_features", staticmethod(lambda corr, th, names, verbose=False: {}))
    monkeypatch.setattr(pimod, "select_features", lambda values, feature_names, **kwargs: feature_names[:1])

    def fake_digraph(X, feature_names, models, connections, root_causes, reciprocity=True, anm_iterations=10, verbose=False):
        edges = [(target, conn) for target, conns in connections.items() for conn in conns]
        g = nx.DiGraph()
        g.add_edges_from(edges)
        return g

    monkeypatch.setattr(pimod.utils, "digraph_from_connected_features", fake_digraph)

    models = DummyModels()
    df = pd.DataFrame({"a": [1.0, 2.0], "b": [2.0, 3.0]})
    pi = pimod.PermutationImportance(models, prog_bar=False, verbose=False)
    pi.fit(df)
    graph = pi.predict(df)

    assert isinstance(graph, nx.DiGraph)
    assert len(graph.edges()) == 2  # each feature connects to one other



"""
Defining a set of classes that represent causal functions/ mechanisms.
"""

import random
import numpy as np
from scipy.stats import bernoulli
from sklearn.mixture import GaussianMixture as GMM
from sklearn.metrics.pairwise import euclidean_distances
from sklearn.gaussian_process import GaussianProcessRegressor


def gmm_cause(n, k=4, p1=2, p2=2, verbose=False):
    g = GMM(k)
    g.means_ = p1 * np.random.randn(k, 1)
    g.covars_ = np.power(abs(p2 * np.random.randn(k, 1) + 1), 2)
    g.weights_ = abs(np.random.rand(k, 1))
    g.weights_ = g.weights_ / sum(g.weights_)

    if(verbose):
        print(f"  GMM cause initialized with {k} components and {n} points")
    return np.random.uniform(-1, 1, n)


class LinearMechanism(object):
    """Linear mechanism, where Effect = alpha*Cause + Noise."""

    def __init__(self, ncauses, points, d=4, noise_coeff=.7, verbose=False):
        """Init the mechanism."""
        super(LinearMechanism, self).__init__()
        self.n_causes = ncauses
        self.points = points

        self.coefflist = []

        for i in range(ncauses):
            self.coefflist.append(random.random())

        self.noise = np.random.randn(points, 1)
        self.d = d

    def __call__(self, causes, verbose=False):
        """Run the mechanism."""
        # Additive only, for now
        effect = np.zeros((self.points, 1))
        # Compute each cause's contribution
        for par in range(causes.shape[1]):
            effect[:, 0] = effect[:, 0] + self.coefflist[par]*causes[:, par]
        effect[:, 0] = effect[:, 0] + self.noise[:, 0]

        return effect


class Polynomial_Mechanism(object):

    def __init__(self, ncauses, points, d=2, noise_coeff=.7, verbose=False):
        """Init the mechanism."""
        super(Polynomial_Mechanism, self).__init__()
        self.n_causes = ncauses
        self.points = points
        self.d = d
        self.polycause = []
        self.verbose = verbose

        for c in range(ncauses):
            self.coefflist = []
            for j in range(self.d + 1):
                self.coefflist.append(random.random())
            self.polycause.append(self.coefflist)

        self.ber = bernoulli.rvs(0.5)
        self.noise = 0.1*np.random.randn(points, 1)
        if self.verbose:
            print(f"Polynomial Mechanism Initialized w {ncauses} causes + {points} p")


    def mechanism(self, x, par):
        list_coeff = self.polycause[par]
        result = np.zeros((self.points, 1))
        for i in range(self.points):
            for j in range(self.d+1):
                result[i, 0] += list_coeff[j]*np.power(x[i], j)
            # XXX
            # ARGH!!!!!!!!!!!!!!!
            # result[i, 0] = min(result[i, 0], 1)
            # result[i, 0] = max(result[i, 0], -1)

        if self.verbose:
            print(f"    Coeffs: ", end="", sep="")
            for j in range(self.d+1):
                print(f"{list_coeff[j]:.2f}, ", end="", sep="")
            print("")
            print(f"    {par}th cause = ", end="", sep="")
            for j in range(self.d+1):
                print(f"{list_coeff[j]:.2f}*x^{j} + ", end="", sep="")
            print(f"noise")
        return result

    def __call__(self, causes, verbose=False):
        """Run the mechanism."""
        if verbose:
            print("  Polynomial Mechanism Called")
        effect = np.zeros((self.points, 1))
        # Compute each cause's contribution
        for par in range(causes.shape[1]):
            effect[:, 0] = effect[:, 0] + self.mechanism(causes[:, par], par)[:, 0]

        # I remove the multiplicative noise since it destroys the effect
        # if(self.ber > 0 and causes.shape[1] > 0):
        #     effect[:, 0] = effect[:, 0] * self.noise[:, 0]
        #     if verbose:
        #         print("    Noise **multiplied** to effect")
        # else:
        effect[:, 0] = effect[:, 0] + self.noise[:, 0]
        if verbose:
            print(f"    Noise added: [", end="")
            for i in range(self.noise.shape[0]):
                print(f"{self.noise[i, 0]:.2f}, ", end="")
            print("]")

        return effect


class SigmoidAM_Mechanism(object):

    def __init__(self, ncauses, points, d=4, noise_coeff=.7, verbose=False):
        """Init the mechanism."""
        super(SigmoidAM_Mechanism, self).__init__()
        self.n_causes = ncauses
        self.points = points

        self.a = np.random.exponential(1/4) + 1
        ber = bernoulli.rvs(0.5)
        self.b = ber * np.random.uniform(-2, -0.5) + (1-ber)*np.random.uniform(0.5, 2)
        self.c = np.random.uniform(-2, 2)
        self.noise = 0.1*np.random.randn(points, 1)

    def mechanism(self, x):
        result = np.\
            zeros((self.points, 1))
        for i in range(self.points):

            result[i, 0] = self.a * self.b * (x[i] + self.c) / (1 + abs(self.b * (x[i] + self.c)))

        return result

    def __call__(self, causes, verbose=False):
        """Run the mechanism."""
        # Additive only
        effect = np.zeros((self.points, 1))
        # Compute each cause's contribution
        for par in range(causes.shape[1]):
            effect[:, 0] = effect[:, 0] + self.mechanism(causes[:, par])[:, 0]

        effect[:, 0] = effect[:, 0] + self.noise[:, 0]

        return effect


class SigmoidMix_Mechanism(object):

    def __init__(self, ncauses, points, d=4, noise_coeff=.7, verbose=False):
        """Init the mechanism."""
        super(SigmoidMix_Mechanism, self).__init__()
        self.n_causes = ncauses
        self.points = points

        self.a = np.random.exponential(1/4) + 1
        ber = bernoulli.rvs(0.5)
        self.b = ber * np.random.uniform(-2, -0.5) + (1-ber)*np.random.uniform(0.5, 2)
        self.c = np.random.uniform(-2, 2)

        self.noise = 0.1*np.random.randn(points, 1)

    def mechanism(self, causes):
        result = np.zeros((self.points, 1))
        for i in range(self.points):
            pre_add_effect = 0
            for c in range(causes.shape[1]):
                pre_add_effect += causes[i, c]
            pre_add_effect += self.noise[i]

            result[i, 0] = self.a * self.b * \
                (pre_add_effect + self.c)/(1 + abs(self.b*(pre_add_effect + self.c)))

        return result

    def __call__(self, causes, verbose=False):
        """Run the mechanism."""
        effect = np.zeros((self.points, 1))
        # Compute each cause's contribution

        effect[:, 0] = self.mechanism(causes)[:, 0]
        return effect


def computeGaussKernel(x):
    xnorm = np.power(euclidean_distances(x, x), 2)
    return np.exp(-xnorm / (2.0))


class GaussianProcessAdd_Mechanism(object):

    def __init__(self, ncauses, points, verbose=False):
        """Init the mechanism."""
        super(GaussianProcessAdd_Mechanism, self).__init__()
        self.n_causes = ncauses
        self.points = points

        self.noise = 0.1*np.random.randn(points, 1)
        self.nb_step = 0

    def mechanism(self, x):
        """Run the mechanism."""
        self.nb_step += 1
        x = np.reshape(x, (x.shape[0], 1))

        if(self.nb_step < 5):
            cov = computeGaussKernel(x)
            mean = np.zeros((1, self.points))[0, :]
            y = np.random.multivariate_normal(mean, cov)
        elif(self.nb_step == 5):
            cov = computeGaussKernel(x)
            mean = np.zeros((1, self.points))[0, :]
            y = np.random.multivariate_normal(mean, cov)
            self.gpr = GaussianProcessRegressor()
            self.gpr.fit(x, y)
            y = self.gpr.predict(x)
        else:
            y = self.gpr.predict(x)

        return y

    def __call__(self, causes, verbose=False):
        """Run the mechanism."""
        # Additive only
        effect = np.zeros((self.points, 1))
        # Compute each cause's contribution
        for par in range(causes.shape[1]):
            effect[:, 0] = effect[:, 0] + self.mechanism(causes[:, par])

        effect[:, 0] = effect[:, 0] + self.noise[:, 0]

        return effect


class GaussianProcessMix_Mechanism(object):

    def __init__(self, ncauses, points, verbose=False):
        """Init the mechanism."""
        super(GaussianProcessMix_Mechanism, self).__init__()
        self.n_causes = ncauses
        self.points = points
        self.noise = 0.1*np.random.randn(points, 1)
        self.nb_step = 0

    def mechanism(self, x):
        self.nb_step += 1
        x = np.reshape(x, (x.shape[0], x.shape[1]))

        if(self.nb_step < 2):
            cov = computeGaussKernel(x)
            mean = np.zeros((1, self.points))[0, :]
            y = np.random.multivariate_normal(mean, cov)
        elif(self.nb_step == 2):
            cov = computeGaussKernel(x)
            mean = np.zeros((1, self.points))[0, :]
            y = np.random.multivariate_normal(mean, cov)
            self.gpr = GaussianProcessRegressor()
            self.gpr.fit(x, y)
            y = self.gpr.predict(x)
        else:
            y = self.gpr.predict(x)

        return y

    def __call__(self, causes, verbose=False):
        """Run the mechanism."""
        effect = np.zeros((self.points, 1))
        # Compute each cause's contribution
        if(causes.shape[1] > 0):
            mix = np.hstack((causes, self.noise))
            effect[:, 0] = self.mechanism(mix)
        else:
            effect[:, 0] = self.mechanism(self.noise)

        return effect


def gaussian_cause(n):
    return np.random.randn(n, 1)[:, 0]


def noise(n, v):
    return v * np.random.rand(1) * np.random.randn(n, 1) + random.sample([2, -2], 1)





import numpy as np
import pandas as pd
import pytest

from causalexplain.explainability import shapley


class DummyModels:
    def __init__(self):
        self.regressor = {"a": object(), "b": object()}


def test_predict_requires_fit():
    estimator = shapley.ShapEstimator(models=DummyModels(), mean_shap_percentile=0.0, prog_bar=False, verbose=False)
    estimator.is_fitted_ = False
    with pytest.raises(ValueError):
        estimator.predict(pd.DataFrame({"a": [1, 2], "b": [3, 4]}))


def test_compute_scaled_shap_threshold():
    estimator = shapley.ShapEstimator(models=DummyModels(), mean_shap_percentile=0.5, prog_bar=False, verbose=False)
    estimator.all_mean_shap_values = np.array([0.0, 1.0, 3.0])
    estimator._compute_scaled_shap_threshold()
    assert estimator.mean_shap_threshold == pytest.approx(1.0)


@pytest.mark.internal
def test_run_selected_shap_explainer_invalid(monkeypatch):
    estimator = shapley.ShapEstimator(models=DummyModels(), explainer="invalid", prog_bar=False, verbose=False)
    with pytest.raises(ValueError):
        estimator._run_selected_shap_explainer("t", object(), np.zeros((1, 1)), np.zeros((1, 1)))


@pytest.mark.internal
def test_compute_discrepancy_outputs_dataclass():
    estimator = shapley.ShapEstimator(models=DummyModels(), prog_bar=False, verbose=False)
    result = estimator._compute_discrepancy(
        x=np.array([0.0, 1.0, 2.0, 3.0]),
        y=np.array([1.0, 2.0, 3.0, 4.0]),
        s=np.array([[0.5], [1.0], [1.5], [2.0]]),
        target_name="t",
        parent_name="p",
    )
    assert isinstance(result, shapley.ShapDiscrepancy)
    assert 0.0 <= result.ks_pvalue <= 1.0


@pytest.mark.internal
def test_compute_discrepancies_builds_matrix():
    estimator = shapley.ShapEstimator(models=DummyModels(), prog_bar=False, verbose=False)
    estimator.feature_names = ["a", "b"]
    estimator.shap_values = {
        "a": np.array([[1.0], [2.0], [3.0], [4.0]]),
        "b": np.array([[4.0], [3.0], [2.0], [1.0]]),
    }
    estimator.is_fitted_ = True
    data = pd.DataFrame({"a": [1, 2, 3, 4], "b": [4, 3, 2, 1]})

    discrepancies = estimator._compute_discrepancies(data)
    assert set(discrepancies.columns) == {"a", "b"}
    assert discrepancies.isna().sum().sum() == 0



import importlib
import sys
import types

import numpy as np


def _load_module_with_stub(monkeypatch, feature_importances):
    def fake_train(X, y, pars=None):
        return {"model": type("M", (), {"feature_importances_": np.array(feature_importances)})()}

    python_pkg = types.ModuleType("Python")
    cam_pkg = types.ModuleType("Python.CAM")
    train_pkg = types.ModuleType("Python.CAM.train_LMboost")
    train_pkg.train_LMboost = fake_train
    cam_pkg.train_LMboost = train_pkg
    python_pkg.CAM = cam_pkg

    monkeypatch.setitem(sys.modules, "Python", python_pkg)
    monkeypatch.setitem(sys.modules, "Python.CAM", cam_pkg)
    monkeypatch.setitem(sys.modules, "Python.CAM.train_LMboost", train_pkg)

    module = importlib.import_module("causalexplain.estimators.cam.selLmBoost")
    monkeypatch.setattr(module, "train_LMboost", fake_train)
    return module


def test_sel_lmboost_selects_top_features(monkeypatch):
    module = _load_module_with_stub(monkeypatch, [0.2, 0.8])

    X = np.array([[1.0, 2.0, 3.0], [1.5, 2.5, 3.5]])
    selection = module.selLmBoost(X, pars={"atLeastThatMuchSelected": 0.1, "atMostThatManyNeighbors": 2}, k=1)

    assert len(selection) == X.shape[1]
    assert selection.count(True) == 2


def test_sel_lmboost_single_variable(monkeypatch):
    module = _load_module_with_stub(monkeypatch, [0.5])
    X = np.ones((2, 1))

    assert module.selLmBoost(X) == []



import pytest

from causalexplain.common import metrics_config


def test_global_metric_types_are_lists_of_strings():
    # Ensure both metric collections are exposed and non-empty.
    assert isinstance(metrics_config.global_metric_types, list)
    assert isinstance(metrics_config.global_nc_metric_types, list)
    assert all(isinstance(m, str) for m in metrics_config.global_metric_types)
    assert all(isinstance(m, str) for m in metrics_config.global_nc_metric_types)
    # Spot-check a couple of known entries to guard against regressions.
    assert "f1" in metrics_config.global_metric_types
    assert "r2" in metrics_config.global_nc_metric_types



"""
This Python version aims to replicate the functionality of the R function.
Here are some key points about the translation:

1. We use NumPy for array operations.
2. Instead of `gam` from R, we use `pygam` library which provides similar
    functionality in Python.
3. The p-values are extracted from the fitted GAM model's statistics.
4. The logic for creating and updating `selVec` is adjusted to work with Python's
    0-based indexing.

Note that this translation assumes that the `pygam` library is installed and imported.
You may need to install it using `pip install pygam`.

Also, be aware that there might be some differences in the exact implementation
details between R's `gam` and Python's `pygam`. You may need to fine-tune the GAM
model creation and fitting process to match the exact behavior of the R version.
"""
import numpy as np

from causalexplain.estimators.cam.train_gam import train_gam


def selGam(X, pars=None, verbose=False, k=None):
    """
    This method selects features based on GAM p-values. It returns a vector
    of selected features whose p-values are less than the cutOffPVal.

    Args:
        X (_type_): _description_
        pars (_type_, optional): _description_. Defaults to None.
        verbose (bool, optional): _description_. Defaults to False.
        k (_type_, optional): _description_. Defaults to None.

    Returns:
        _type_: _description_
    """
    if pars is None:
        pars = {'cutOffPVal': 0.001, 'numBasisFcts': 10}

    X = np.asarray(X)
    p = X.shape

    if p[1] > 1:
        selVec = [False] * p[1]
        X_without_k = np.delete(X, k-1, axis=1)
        y = X[:, k-1]

        # Use train_gam function instead of directly creating and fitting GAM
        gam_result = train_gam(X_without_k, y, pars=pars, verbose=verbose)

        # Extract p-values from the gam_result.
        # PyGAM returns the p-values of all predictors, followed by the p-value of
        # the intercept, which we don't need.
        pValVec = gam_result['p_values']
        pValVec = pValVec[:-1]


        if len(pValVec) != len(selVec) - 1:
            print("This should never happen (function selGam).")

        selVec_without_k = [p < pars['cutOffPVal'] for p in pValVec]
        selVec[:k] = selVec_without_k[:k]
        selVec[k+1:] = selVec_without_k[k:]
        if verbose:
            print(f". . . . . SelGAM(k={k})")
            print(f". . . . . . Vector of p-values: {pValVec}")
            print(f". . . . . . Selected indices: {selVec_without_k}")

    else:
        selVec = []

    return selVec



import networkx as nx
import pandas as pd
import pytest

from causalexplain.estimators.pc.pc import PC
from causalexplain.estimators.pc import pc as pc_module
from causalexplain.estimators.pc.pdag import PDAG


def test_fit_validates_variant_and_ci_test():
    pc = PC(name="demo", variant="invalid")
    with pytest.raises(ValueError):
        pc.fit(pd.DataFrame({"A": [1]}))

    pc = PC(name="demo", ci_test="bad")
    with pytest.raises(ValueError):
        pc.fit(pd.DataFrame({"A": [1]}))


def test_fit_returns_skeleton_and_dag(monkeypatch):
    data = pd.DataFrame({"A": [1, 2], "B": [2, 3]})
    pc = PC(name="demo", variant="stable", return_type="skeleton")

    skeleton = nx.Graph()
    skeleton.add_edge("A", "B")
    monkeypatch.setattr(pc, "build_skeleton", lambda **kwargs: (skeleton, {}))

    skel, sep_sets = pc.fit(data)
    assert skel is skeleton and sep_sets == {}

    pc2 = PC(name="demo", return_type="dag")
    monkeypatch.setattr(pc2, "build_skeleton", lambda **kwargs: (skeleton, {frozenset({"A", "B"}): set()}))

    class DummyPDAG(PDAG):
        def __init__(self):
            super().__init__(directed_ebunch=[("A", "B")], undirected_ebunch=[])

        def to_dag(self):
            return nx.DiGraph([("A", "B")])

    monkeypatch.setattr(pc_module.PC, "skeleton_to_pdag", staticmethod(lambda sk, sep: DummyPDAG()))
    dag = pc2.fit(data)
    assert isinstance(dag, nx.DiGraph)
    assert pc2.is_fitted_


def test_fit_predict_calls_evaluate_graph(monkeypatch):
    data = pd.DataFrame({"A": [1, 2], "B": [2, 3]})
    pc = PC(name="demo", return_type="dag")

    skeleton = nx.Graph()
    skeleton.add_edge("A", "B")
    monkeypatch.setattr(pc, "build_skeleton", lambda **kwargs: (skeleton, {frozenset({"A", "B"}): set()}))
    monkeypatch.setattr(pc_module.PC, "skeleton_to_pdag", staticmethod(lambda sk, sep: PDAG(directed_ebunch=[("A", "B")], undirected_ebunch=[])))

    metrics_called = {}

    def fake_evaluate_graph(ref, dag, feature_names=None):
        metrics_called["called"] = True
        return "metrics"

    monkeypatch.setattr(pc_module, "evaluate_graph", fake_evaluate_graph)

    ref_graph = nx.DiGraph()
    ref_graph.add_node("A")
    result = pc.fit_predict(data, data, ref_graph=ref_graph)
    assert isinstance(result, nx.DiGraph)
    assert metrics_called["called"]


def test_skeleton_to_pdag_orients_v_structure():
    skeleton = nx.Graph()
    skeleton.add_edge("X", "Z")
    skeleton.add_edge("Y", "Z")
    separating_sets = {frozenset({"X", "Y"}): set()}

    pdag = PC.skeleton_to_pdag(skeleton, separating_sets)
    assert ("X", "Z") in pdag.directed_edges
    assert ("Y", "Z") in pdag.directed_edges
    assert ("Z", "X") not in pdag.directed_edges


def test_fit_invalid_return_type(monkeypatch):
    data = pd.DataFrame({"A": [1, 2], "B": [2, 3]})
    pc = PC(name="demo", return_type="unknown")
    skeleton = nx.Graph()
    skeleton.add_edge("A", "B")
    monkeypatch.setattr(pc, "build_skeleton", lambda **kwargs: (skeleton, {frozenset({"A", "B"}): set()}))
    monkeypatch.setattr(pc_module.PC, "skeleton_to_pdag", staticmethod(lambda sk, sep: PDAG(directed_ebunch=[("A", "B")], undirected_ebunch=[])))

    with pytest.raises(ValueError):
        pc.fit(data)



"""
This module contains the GraphDiscovery class which is responsible for
creating, fitting, and evaluating causal discovery experiments.
"""
import os
import pickle
import pandas as pd
import matplotlib.pyplot as plt
from typing import List, Tuple

from causalexplain.common import (
    DEFAULT_REGRESSORS,
    utils,
)
from causalexplain.common import plot
from causalexplain.common.notebook import Experiment
from causalexplain.metrics.compare_graphs import evaluate_graph


class GraphDiscovery:
    def __init__(
        self,
        experiment_name: str = None,
        model_type: str = 'rex',
        csv_filename: str = None,
        true_dag_filename: str = None,
        verbose: bool = False,
        seed: int = 42
    ) -> None:
        """
        Initializes a new instance of the GraphDiscovery class.

        Args:
            experiment_name (str, optional): The name of the experiment.
            model_type (str, optional): The type of model to use. Valid options
                are: 'rex', 'pc', 'fci', 'ges', 'lingam', 'cam', 'notears'.
            csv_filename (str, optional): The filename of the CSV file containing
                the data.
            true_dag_filename (str, optional): The filename of the DOT file
                containing the true causal graph.
            verbose (bool, optional): Whether to print verbose output.
            seed (int, optional): The random seed for reproducibility.
        """
        # Normalize empty/whitespace strings to None
        experiment_name = experiment_name.strip() if isinstance(
            experiment_name, str) else experiment_name
        experiment_name = None if experiment_name == "" else experiment_name
        csv_filename = csv_filename.strip() if isinstance(
            csv_filename, str) else csv_filename
        csv_filename = None if csv_filename == "" else csv_filename

        if (experiment_name is None and csv_filename is not None) or \
                (experiment_name is not None and csv_filename is None):
            raise ValueError(
                f"Both 'experiment_name' and 'csv_filename' must be provided together, "
                f"or none of them. Got experiment_name='{experiment_name}', "
                f"csv_filename='{csv_filename}'")
        elif experiment_name is None and csv_filename is None:
            self.experiment_name = None
            self.estimator = 'rex'
            self.csv_filename = None
            self.dot_filename = None
            self.verbose = False
            self.seed = 42
            return

        self.experiment_name = experiment_name
        self.estimator = model_type
        self.csv_filename = csv_filename
        self.dot_filename = true_dag_filename
        self.verbose = verbose
        self.seed = seed

        self.dataset_path = os.path.dirname(csv_filename)
        self.output_path = os.getcwd()
        self.trainer = {}

        # Read the reference graph
        if true_dag_filename is not None:
            self.ref_graph = utils.graph_from_dot_file(true_dag_filename)
        else:
            self.ref_graph = None

        # assert that the data file exists
        if not os.path.exists(csv_filename):
            raise FileNotFoundError(f"Data file {csv_filename} not found")
        self.dataset_name = os.path.splitext(os.path.basename(csv_filename))[0]

        # Read the column names of the data.
        data = pd.read_csv(csv_filename)
        self.data_columns = list(data.columns)
        del data

        if self.estimator == 'rex':
            self.regressors = DEFAULT_REGRESSORS
        else:
            self.regressors = [self.estimator]

    def create_experiments(self) -> dict:
        """
        Create an Experiment object for each regressor.

        Args:
            dataset_name (str): Name of the dataset
            true_dag (str): Path to the true DAG DOT file
            regressors (list): List of regressor types to create experiments for
            dataset_path (str): Path to the input dataset
            output_path (str): Path for output files

        Returns:
            dict: A dictionary of Experiment objects
        """
        self.trainer = {}
        for model_type in self.regressors:
            trainer_name = f"{self.dataset_name}_{model_type}"
            self.trainer[trainer_name] = Experiment(
                experiment_name=self.dataset_name,
                csv_filename=self.csv_filename,
                dot_filename=self.dot_filename,
                model_type=model_type,
                input_path=self.dataset_path,
                output_path=self.output_path,
                verbose=False)

        return self.trainer

    def fit_experiments(
        self,
        hpo_iterations: int = None,
        bootstrap_iterations: int = None,
        prior: List[List[str]] = None,
        **kwargs
    ) -> None:
        """
        Fit the Experiment objects.

        Args:
            trainer (dict): A dictionary of Experiment objects
            estimator (str): The estimator to use ('rex' or other)
            verbose (bool, optional): Whether to print verbose output.
                Defaults to False.
            hpo_iterations (int, optional): Number of HPO trials for REX.
                Defaults to None.
            bootstrap_iterations (int, optional): Number of bootstrap trials
                for REX. Defaults to None.
        """
        if self.estimator == 'rex':
            xargs = {
                'verbose': self.verbose,
                'hpo_n_trials': hpo_iterations,
                'bootstrap_trials': bootstrap_iterations,
                # 'prior': prior
            }
        else:
            xargs = {
                'verbose': self.verbose
            }

        # Combine the arguments
        xargs.update(kwargs)

        for trainer_name, experiment in self.trainer.items():
            if not trainer_name.endswith("_rex"):
                experiment.fit_predict(estimator=self.estimator, **xargs)

    def combine_and_evaluate_dags(self, prior: List[List[str]] = None) -> Experiment:
        """
        Retrieve the DAG from the Experiment objects.

        Args:
            prior (List[List[str]], optional): The prior to use for ReX.
                Defaults to None.

        Returns:
            Experiment: The experiment object with the final DAG
        """
        if self.estimator != 'rex':
            trainer_key = f"{self.dataset_name}_{self.estimator}"
            estimator_obj = getattr(self.trainer[trainer_key], self.estimator)
            self.trainer[trainer_key].dag = estimator_obj.dag
            if self.ref_graph is not None and self.data_columns is not None:
                self.trainer[trainer_key].metrics = evaluate_graph(
                    self.ref_graph, estimator_obj.dag, self.data_columns)
            else:
                self.trainer[trainer_key].metrics = None

            self.dag = self.trainer[trainer_key].dag
            self.metrics = self.trainer[trainer_key].metrics
            return self.trainer[trainer_key]

        # For ReX, we need to combine the DAGs. Hardcoded for now to combine
        # the first and second DAGs
        estimator1 = getattr(self.trainer[list(self.trainer.keys())[0]], 'rex')
        estimator2 = getattr(self.trainer[list(self.trainer.keys())[1]], 'rex')
        _, _, dag, _ = utils.combine_dags(
            estimator1.dag, estimator2.dag,
            discrepancies=estimator1.shaps.shap_discrepancies,
            prior=prior
        )

        # Create a new Experiment object for the combined DAG
        new_trainer = f"{self.dataset_name}_rex"
        if new_trainer in self.trainer:
            del self.trainer[new_trainer]
        self.trainer[new_trainer] = Experiment(
            experiment_name=self.dataset_name,
            model_type='rex',
            input_path=self.dataset_path,
            output_path=self.output_path,
            verbose=False)

        # Set the DAG and evaluate it
        self.trainer[new_trainer].ref_graph = self.ref_graph
        self.trainer[new_trainer].dag = dag
        if self.ref_graph is not None and self.data_columns is not None:
            self.trainer[new_trainer].metrics = evaluate_graph(
                self.ref_graph, dag, self.data_columns)
        else:
            self.trainer[new_trainer].metrics = None

        self.dag = self.trainer[new_trainer].dag
        self.metrics = self.trainer[new_trainer].metrics
        return self.trainer[new_trainer]

    def run(
            self,
            hpo_iterations: int = None,
            bootstrap_iterations: int = None,
            prior: List[List[str]] = None,
            **kwargs):
        """
        Run the experiment.

        Args:
            hpo_iterations (int, optional): Number of HPO trials for REX.
                Defaults to None.
            bootstrap_iterations (int, optional): Number of bootstrap trials
                for REX. Defaults to None.
        """
        self.create_experiments()
        self.fit_experiments(
            hpo_iterations, bootstrap_iterations, prior, **kwargs)
        self.combine_and_evaluate_dags(prior=prior)

    def save(self, full_filename_path: str) -> None:
        """
        Save the model as an Experiment object.

        Args:
            full_filename_path (str): A full path where to save the model,
                including the filename.
        """
        assert self.trainer, "No trainer to save"
        assert full_filename_path, "No output path specified"

        full_dir_path = os.path.dirname(full_filename_path)
        # Check only if not local dir
        if full_dir_path != "." and full_dir_path != "":
            assert os.path.exists(full_dir_path), \
                f"Output directory {full_dir_path} does not exist"
        else:
            full_dir_path = os.getcwd()

        saved_as = utils.save_experiment(
            os.path.basename(full_filename_path), full_dir_path,
            self.trainer, overwrite=False)
        print(f"Saved model as: {saved_as}", flush=True)

    def load(self, model_path: str) -> Experiment:
        """
        Load the model from a pickle file.

        Args:
            model_path (str): Path to the pickle file containing the model

        Returns:
            Experiment: The loaded Experiment object
        """
        with open(model_path, 'rb') as f:
            self.trainer = pickle.load(f)
            print(f"Loaded model from: {model_path}", flush=True)

        # Set the dag and metrics
        self.dag = self.trainer[list(self.trainer.keys())[-1]].dag
        self.metrics = self.trainer[list(self.trainer.keys())[-1]].metrics
        return self.trainer

    def printout_results(self, graph, metrics):
        """
        This method prints the DAG to stdout in hierarchical order.

        Parameters:
        -----------
        dag : nx.DiGraph
            The DAG to be printed.
        """
        if len(graph.edges()) == 0:
            print("Empty graph")
            return

        print("Resulting Graph:\n---------------")

        def dfs(node, visited, indent=""):
            if node in visited:
                return  # Avoid revisiting nodes
            visited.add(node)

            # Print edges for this node
            for neighbor in graph.successors(node):
                print(f"{indent}{node} -> {neighbor}")
                dfs(neighbor, visited, indent + "  ")

        visited = set()

        # Start traversal from all nodes without predecessors (roots)
        for node in graph.nodes:
            if graph.in_degree(node) == 0:
                dfs(node, visited)

        # Handle disconnected components (not reachable from any "root")
        for node in graph.nodes:
            if node not in visited:
                dfs(node, visited)

        if metrics is not None:
            print("\nGraph Metrics:\n-------------")
            print(metrics)


    def export(self, output_file: str) -> str:
        """
        This method exports the DAG to a DOT file.

        Parameters:
        -----------
        dag : nx.DiGraph
            The DAG to be exported.
        output_file : str
            The path to the output DOT file.

        Returns:
        --------
        str
            The path to the output DOT file.
        """
        saved_as = utils.graph_to_dot_file(
            self.trainer[list(self.trainer.keys())[-1]].dag, output_file)

        return saved_as

    def plot(
        self,
        show_metrics: bool = False,
        show_node_fill: bool = True,
        title: str = None,
        ax: plt.Axes = None,
        figsize: Tuple[int, int] = (5, 5),
        dpi: int = 75,
        save_to_pdf: str = None,
        layout: str = 'dot',
        **kwargs
    ):
        """
        This method plots the DAG using networkx and matplotlib.

        Parameters:
        -----------
        show_metrics : bool, optional
            Whether to show the metrics on the plot. Defaults to False.
        show_node_fill : bool, optional
            Whether to fill the nodes with color. Defaults to True.
        title : str, optional
            The title of the plot. Defaults to None.
        ax : plt.Axes, optional
            The matplotlib axes to plot on. Defaults to None.
        figsize : Tuple[int, int], optional
            The size of the plot. Defaults to (5, 5).
        dpi : int, optional
            The DPI of the plot. Defaults to 75.
        save_to_pdf : str, optional
            The path to save the plot as a PDF. Defaults to None.
        layout : str, optional
            The layout to use for the plot. Defaults to 'dot'. Other option
            is 'circular'.
        """
        model = self.trainer[list(self.trainer.keys())[-1]]
        if model.ref_graph is not None:
            ref_graph = model.ref_graph
        else:
            ref_graph = None
        plot.dag(
            graph=model.dag, reference=ref_graph, show_metrics=show_metrics,
            show_node_fill=show_node_fill, title=title, ax=ax,
            figsize=figsize, dpi=dpi, save_to_pdf=save_to_pdf, layout=layout,
            **kwargs)

    @property
    def model(self):
        return self.trainer[list(self.trainer.keys())[-1]]



import numpy as np

from causalexplain.estimators.notears import loss


def test_least_squares_loss_and_grad():
    data = np.array([[1.0, 2.0], [3.0, 4.0]])
    W = np.array([[0.0, 0.5], [0.0, 0.0]])
    cov = np.cov(data.T)
    d, n = data.shape[1], data.shape[0]

    computed_loss = loss.least_squares_loss(W, data, cov, d, n)
    expected_loss = 0.5 / n * np.linalg.norm(data - data @ W, ord="fro") ** 2
    assert np.isclose(computed_loss, expected_loss)

    grad = loss.least_squares_loss_grad(W, data, cov, d, n)
    expected_grad = (-1.0 / n) * data.T @ (data - data @ W)
    assert np.allclose(grad, expected_grad)


def test_least_squares_cov_variants():
    data = np.array([[1.0, 0.0]])
    W = np.array([[0.0]])
    cov = np.array([[2.0]])
    d, n = 1, 1

    loss_cov = loss.least_squares_loss_cov(W, data, cov, d, n)
    expected_loss = 0.5 * np.trace((np.eye(d) - W).T @ cov.T @ (np.eye(d) - W))
    assert np.isclose(loss_cov, expected_loss)

    grad_cov = loss.least_squares_loss_cov_grad(W, data, cov, d, n)
    expected_grad = (-0.5) * ((cov + cov.T) @ (np.eye(d) - W))
    assert np.allclose(grad_cov, expected_grad)



import types

import pandas as pd
import pytest

from causalexplain.common import notebook


class DummyEstimator:
    def __init__(self, name=None, **kwargs):
        self.name = name
        self.kwargs = kwargs
        self.fit_args = None
        self.predict_args = None

    def fit(self, *args, **kwargs):
        self.fit_args = (args, kwargs)
        return self

    def predict(self, *args, **kwargs):
        self.predict_args = (args, kwargs)
        return self

    def fit_predict(self, *args, **kwargs):
        self.fit_args = (args, kwargs)
        return self


@pytest.fixture
def sample_csv(tmp_path):
    path = tmp_path / "data.csv"
    df = pd.DataFrame({"a": [1.0, 2.0, 3.0], "b": [4.0, 5.0, 6.0]})
    df.to_csv(path, index=False)
    return path


def test_check_model_type_rejects_unknown():
    exp = notebook.Experiment.__new__(notebook.Experiment)
    # Initialize BaseExperiment portion without touching filesystem
    notebook.BaseExperiment.__init__(
        exp,
        input_path="",
        output_path="",
        train_anyway=False,
        save_anyway=False,
        scale=False,
        train_size=0.9,
        random_state=42,
        verbose=False,
    )
    with pytest.raises(ValueError):
        exp._check_model_type("unsupported")


def test_prepare_experiment_input_scaling_and_graph(monkeypatch, sample_csv, tmp_path):
    # Return a sentinel graph without hitting filesystem
    sentinel_graph = object()
    monkeypatch.setattr(notebook.utils, "graph_from_dot_file", lambda path: sentinel_graph)
    exp = notebook.Experiment(
        experiment_name="exp",
        csv_filename=str(sample_csv),
        dot_filename=str(tmp_path / "graph.dot"),
        model_type="nn",
        input_path=str(tmp_path),
        output_path=str(tmp_path),
        train_size=2 / 3,
        random_state=0,
        verbose=False,
    )
    assert exp.ref_graph is sentinel_graph
    # train/test split respects size and scaling flag default (False)
    assert len(exp.train_data) + len(exp.test_data) == len(exp.data)


def test_create_estimator_unknown(monkeypatch, sample_csv, tmp_path):
    monkeypatch.setattr(notebook.utils, "graph_from_dot_file", lambda _: None)
    exp = notebook.Experiment(
        experiment_name="exp",
        csv_filename=str(sample_csv),
        dot_filename=str(tmp_path / "graph.dot"),
        model_type="nn",
        input_path=str(tmp_path),
        output_path=str(tmp_path),
    )
    assert exp.create_estimator("nonexistent", name="x") is None


def test_fit_and_predict_use_created_estimator(monkeypatch, sample_csv, tmp_path):
    monkeypatch.setattr(notebook.utils, "graph_from_dot_file", lambda _: None)
    exp = notebook.Experiment(
        experiment_name="exp",
        csv_filename=str(sample_csv),
        dot_filename=str(tmp_path / "graph.dot"),
        model_type="nn",
        input_path=str(tmp_path),
        output_path=str(tmp_path),
    )
    monkeypatch.setattr(exp, "create_estimator", lambda estimator_name, name, **kwargs: DummyEstimator(name=name, **kwargs))
    exp.fit(estimator_name="rex", pipeline="p")
    assert isinstance(getattr(exp, "rex"), DummyEstimator)
    assert exp.is_fitted is True
    # Predict delegates to estimator stored under estimator_name
    exp.predict(estimator="rex")
    assert getattr(exp, "rex").predict_args[0][0] is exp.train_data


def test_fit_predict_handles_missing_estimator(monkeypatch, sample_csv, tmp_path):
    monkeypatch.setattr(notebook.utils, "graph_from_dot_file", lambda _: None)
    exp = notebook.Experiment(
        experiment_name="exp",
        csv_filename=str(sample_csv),
        dot_filename=str(tmp_path / "graph.dot"),
        model_type="nn",
        input_path=str(tmp_path),
        output_path=str(tmp_path),
    )
    monkeypatch.setattr(exp, "create_estimator", lambda *_, **__: None)
    with pytest.raises(ValueError):
        exp.fit_predict(estimator="unknown")


def test_load_and_save_roundtrip(monkeypatch, sample_csv, tmp_path):
    # Use a dummy estimator object and simulate persistence helpers
    dummy = DummyEstimator()
    monkeypatch.setattr(notebook, "Rex", DummyEstimator)
    monkeypatch.setattr(notebook.utils, "graph_from_dot_file", lambda _: None)
    exp = notebook.Experiment(
        experiment_name="exp",
        csv_filename=str(sample_csv),
        dot_filename=str(tmp_path / "graph.dot"),
        model_type="nn",
        input_path=str(tmp_path),
        output_path=str(tmp_path),
    )

    saved = {}
    monkeypatch.setattr(
        notebook.utils,
        "save_experiment",
        lambda name, folder, obj, overwrite=False: saved.update({"name": name, "folder": folder, "obj": obj}) or "path",
    )
    monkeypatch.setattr(
        notebook.utils,
        "load_experiment",
        lambda name, folder: dummy,
    )
    # Pretend estimator is already fitted
    exp.estimator_name = "rex"
    setattr(exp, "rex", dummy)

    path = exp.save(overwrite=True)
    assert path == "path"
    assert saved["obj"] is dummy

    loaded = exp.load()
    assert loaded.estimator_name == "rex"
    assert loaded.estimator is dummy



"""
This module builds the causal graph based on the informacion that we derived
from the SHAP values. The main idea is to use the SHAP values to compute the
discrepancy between the SHAP values and the target values. This discrepancy
is then used to build the graph.
"""

# pylint: disable=E1101:no-member, W0201:attribute-defined-outside-init, W0511:fixme
# pylint: disable=C0103:invalid-name, R0902:too-many-instance-attributes
# pylint: disable=C0116:missing-function-docstring
# pylint: disable=R0913:too-many-arguments, W0621:redefined-outer-name
# pylint: disable=R0914:too-many-locals, R0915:too-many-statements
# pylint: disable=W0106:expression-not-assigned, R1702:too-many-branches

import inspect
import math
import multiprocessing
from collections import defaultdict
from dataclasses import dataclass
from functools import partial
from multiprocessing import get_context
from typing import List, Union, Optional

import colorama
import networkx as nx
import numpy as np
import pandas as pd
import shap
import statsmodels.api as sm
import statsmodels.stats.api as sms
import torch
from matplotlib import pyplot as plt
from matplotlib.ticker import FormatStrFormatter
from mlforge.progbar import ProgBar  # type: ignore
from scipy.stats import kstest, spearmanr
from sklearn.base import BaseEstimator
from sklearn.discriminant_analysis import StandardScaler
from sklearn.metrics import r2_score
from sklearn.model_selection import train_test_split

from ..independence.feature_selection import select_features
from ..common import utils

RED = colorama.Fore.RED
GREEN = colorama.Fore.GREEN
GRAY = colorama.Fore.LIGHTBLACK_EX
RESET = colorama.Style.RESET_ALL

AnyGraph = Union[nx.DiGraph, nx.Graph]
K = 180.0 / math.pi


@dataclass
class ShapDiscrepancy:
    """
    A class representing the discrepancy between the SHAP value and the parent
    value for a given feature.

    Attributes:
        - target (str): The name of the target feature.
        - parent (str): The name of the parent feature.
        - shap_heteroskedasticity (bool): Whether the SHAP value exhibits
            heteroskedasticity.
        - parent_heteroskedasticity (bool): Whether the parent value exhibits
            heteroskedasticity.
        - shap_p_value (float): The p-value for the SHAP value.
        - parent_p_value (float): The p-value for the parent value.
        - shap_model (sm.regression.linear_model.RegressionResultsWrapper): The
            regression model for the SHAP value.
        - parent_model (sm.regression.linear_model.RegressionResultsWrapper): The
            regression model for the parent value.
        - shap_discrepancy (float): The discrepancy between the SHAP value and the
            parent value.
        - shap_correlation (float): The correlation between the SHAP value and the
            parent value.
        - shap_gof (float): The goodness of fit for the SHAP value.
        - ks_pvalue (float): The p-value for the Kolmogorov-Smirnov test.
        - ks_result (str): The result of the Kolmogorov-Smirnov test.
    """
    target: str
    parent: str
    shap_heteroskedasticity: bool
    parent_heteroskedasticity: bool
    shap_p_value: float
    parent_p_value: float
    shap_model: sm.regression.linear_model.RegressionResultsWrapper
    parent_model: sm.regression.linear_model.RegressionResultsWrapper
    shap_discrepancy: float
    shap_correlation: float
    shap_gof: float
    ks_pvalue: float
    ks_result: str


class ShapEstimator(BaseEstimator):
    """
    A class for computing SHAP values and building a causal graph from them.

    Parameters
    ----------
    explainer : str, default="explainer"
        The SHAP explainer to use. Possible values are "kernel", "gradient", and
        "explainer".
    models : BaseEstimator, default=None
        The models to use for computing SHAP values. If None, a linear regression
        model is used for each feature.
    correlation_th : float, default=None
        The correlation threshold to use for removing highly correlated features.
    mean_shap_percentile : float, default=0.8
        The percentile threshold for selecting features based on their mean SHAP value.
    iters : int, default=20
        The number of iterations to use for the feature selection method.
    reciprocity : bool, default=False
        Whether to enforce reciprocity in the causal graph.
    min_impact : float, default=1e-06
        The minimum impact threshold for selecting features.
    exhaustive : bool, default=False
        Whether to use the exhaustive (recursive) method for selecting features.
        If this is True, the threshold parameter must be provided, and the
        clustering is performed until remaining values to be clustered are below
        the given threshold.
    threshold : float, default=None
        The threshold to use when exhaustive is True. If None, exception is raised.
    on_gpu : bool, default=False
        Whether to use the GPU for computing SHAP values.
    verbose : bool, default=False
        Whether to print verbose output.
    prog_bar : bool, default=True
        Whether to show a progress bar.
    silent : bool, default=False
        Whether to suppress all output.
    """

    device = utils.select_device("cpu")
    explainer = "explainer"
    models = None
    correlation_th = None
    mean_shap_percentile = 0.8
    shap_discrepancies = None
    iters = 20
    reciprocity = False
    min_impact = 1e-06
    exhaustive = False
    parallel_jobs = 0
    on_gpu = False
    prog_bar = True
    verbose = False
    silent = False


    def __init__(
            self,
            explainer: str = "explainer",
            models: Optional[BaseEstimator] = None,
            correlation_th: Optional[float] = None,
            mean_shap_percentile: float = 0.8,
            iters: int = 20,
            reciprocity: bool = False,
            min_impact: float = 1e-06,
            exhaustive: bool = False,
            parallel_jobs: int = 0,
            on_gpu: bool = False,
            verbose: bool = False,
            prog_bar: bool = True,
            silent: bool = False):
        """
        Initialize the ShapEstimator object.

        Parameters
        ----------
        explainer : str, default="explainer"
            The SHAP explainer to use. Possible values are "kernel", "gradient", and
            "explainer".
        models : BaseEstimator, default=None
            The models to use for computing SHAP values. If None, a linear regression
            model is used for each feature.
        correlation_th : float, default=None
            The correlation threshold to use for removing highly correlated features.
        mean_shap_percentile : float, default=0.8
            The percentile threshold for selecting features based on their
            mean SHAP value.
        iters : int, default=20
            The number of iterations to use for the feature selection method.
        reciprocity : bool, default=False
            Whether to enforce reciprocity in the causal graph.
        min_impact : float, default=1e-06
            The minimum impact threshold for selecting features.
        exhaustive : bool, default=False
            Whether to use the exhaustive (recursive) method for selecting features.
            If this is True, the threshold parameter must be provided, and the
            clustering is performed until remaining values to be clustered are below
            the given threshold.
        threshold : float, default=None
            The threshold to use when exhaustive is True. If None, exception is raised.
        on_gpu : bool, default=False
            Whether to use the GPU for computing SHAP values.
        verbose : bool, default=False
            Whether to print verbose output.
        prog_bar : bool, default=True
            Whether to show a progress bar.
        silent : bool, default=False
            Whether to suppress all output.
        """
        self.explainer = explainer
        self.models = models
        self.correlation_th = correlation_th
        self.mean_shap_percentile = mean_shap_percentile
        self.iters = iters
        self.reciprocity = reciprocity
        self.min_impact = min_impact
        self.exhaustive = exhaustive
        self.parallel_jobs = parallel_jobs
        self.on_gpu = on_gpu
        self.verbose = verbose
        self.prog_bar = prog_bar
        self.silent = silent

        self._fit_desc = f"Running SHAP explainer ({self.explainer})"
        self._pred_desc = "Building graph skeleton"

    def __str__(self):
        return utils.stringfy_object(self)

    # Define the process_target function at the top level
    @staticmethod
    def _shap_fit_target_variable(
        target_name,
        models,
        X_train,
        X_test,
        feature_names,
        on_gpu,
        verbose,
        run_selected_shap_explainer_func,
    ):
        """
        Process a single target in the SHAP fit process.
        """
        # Get the model and the data (tensor form)
        if hasattr(models.regressor[target_name], "model"):
            model = models.regressor[target_name].model
            model = model.cuda() if on_gpu else model.cpu()
        else:
            model = models.regressor[target_name]

        X_train_target = X_train.drop(target_name, axis=1).values
        X_test_target = X_test.drop(target_name, axis=1).values

        # Run the selected SHAP explainer
        shap_values_target = run_selected_shap_explainer_func(
            target_name, model, X_train_target, X_test_target
        )

        # Create the order list of features, in decreasing mean SHAP value
        feature_order_target = np.argsort(
            np.sum(np.abs(shap_values_target), axis=0))
        shap_mean_values_target = np.abs(shap_values_target).mean(0)

        # Optionally, print verbose output
        if verbose:
            print(f"  Feature order for '{target_name}' {feature_order_target}")
            print(f"  Target({target_name}) -> ", end="")
            srcs = [src for src in feature_names if src != target_name]
            for i in range(len(shap_mean_values_target)):
                print(f"{srcs[i]}:{shap_mean_values_target[i]:.3f};", end="")
            print()

        # Return results
        return target_name, shap_values_target, feature_order_target, \
            shap_mean_values_target


    def fit(self, X):
        """
        Fit the ShapleyExplainer model to the given dataset.

        Parameters:
        - X: The input dataset.

        Returns:
        - self: The fitted ShapleyExplainer model.
        """
        assert self.models is not None, "shap.models must be set"

        self.feature_names = list(self.models.regressor.keys())  # type: ignore
        self.shap_explainer = {}
        self.shap_values = {}
        self.shap_mean_values = {}
        self.feature_order = {}
        self.all_mean_shap_values = np.empty((0,), dtype=np.float16)

        # Initialize the progress bar
        if self.prog_bar and not self.verbose:
            caller_name = self._get_method_caller_name()
            pbar_name = f"({caller_name}) SHAP_fit"
            pbar = ProgBar().start_subtask(pbar_name, len(self.feature_names))
        else:
            pbar = None

        self.X_train, self.X_test = train_test_split(
            X, test_size=min(0.2, 250 / len(X)), random_state=42
        )

        # Prepare arguments for partial function
        partial_process_target = partial(
            ShapEstimator._shap_fit_target_variable,
            models=self.models,
            X_train=self.X_train,
            X_test=self.X_test,
            feature_names=self.feature_names,
            on_gpu=self.on_gpu,
            verbose=self.verbose,
            run_selected_shap_explainer_func=self._run_selected_shap_explainer,
        )

        if self.parallel_jobs != 0:
            if self.parallel_jobs == -1:
                nr_processes = min(multiprocessing.cpu_count(), len(self.feature_names))
            else:
                nr_processes = min(self.parallel_jobs, multiprocessing.cpu_count())

            # Use 'spawn' start method to avoid semaphore leaks
            with get_context('spawn').Pool(processes=nr_processes) as pool:
                results = []
                for result in pool.imap_unordered(partial_process_target, self.feature_names):
                    results.append(result)
                    if pbar:
                        pbar.update_subtask(pbar_name, len(results))  # type: ignore
                pbar.remove(pbar_name) if pbar else None            # type: ignore
                pbar = None
                pool.close()
                pool.join()
        else:
            # Sequential processing
            results = []
            for target_name in self.feature_names:
                result = partial_process_target(target_name)
                results.append(result)
                if pbar:
                    pbar.update_subtask(pbar_name, len(results))  # type: ignore

        # Collect results
        for target_name, shap_values_target, feature_order_target, shap_mean_values_target in results:
            self.shap_values[target_name] = shap_values_target
            self.feature_order[target_name] = feature_order_target
            self.shap_mean_values[target_name] = shap_mean_values_target

            # If dimensions of all_mean_shap_values is not (m, 1), reshape it
            if len(self.all_mean_shap_values.shape) == 1:
                self.all_mean_shap_values = self.all_mean_shap_values.reshape(-1, 1)

            if len(shap_mean_values_target.shape) == 1:
                shap_mean_values_target = shap_mean_values_target.reshape(-1, 1)

            self.all_mean_shap_values = np.concatenate(
                (self.all_mean_shap_values, shap_mean_values_target)
            )

        if pbar:
            pbar.remove(pbar_name)  # type: ignore

        self.all_mean_shap_values = self.all_mean_shap_values.flatten()
        self._compute_scaled_shap_threshold()

        self.is_fitted_ = True

        return self

    def _compute_scaled_shap_threshold(self):
        """
        Compute the scaled SHAP threshold based on the given percentile.
        If the percentile is 0.0 or None, then the threshold is set to 0.0.
        """
        if self.mean_shap_percentile:
            self.mean_shap_threshold = np.quantile(
                self.all_mean_shap_values, self.mean_shap_percentile)
        else:
            self.mean_shap_threshold = 0.0

    def _run_selected_shap_explainer(self, target_name, model, X_train, X_test):
        """
        Run the selected SHAP explainer, according to the given parameters.

        Parameters
        ----------
        target_name : str
            The name of the target feature.
        model : torch.nn.Module
            The model for the given target.
        X_train : PyTorch.Tensor object
            The training data.
        X_test : PyTorch.Tensor object
            The testing data.

        Returns
        -------
        shap_values : np.ndarray
            The SHAP values for the given target.
        """
        if self.explainer == "kernel":
            self.shap_explainer[target_name] = shap.KernelExplainer(
                model.predict, X_train)
            shap_values = self.shap_explainer[target_name].\
                shap_values(X_test)[0]
        elif self.explainer == "gradient":
            X_train_tensor = torch.from_numpy(X_train).float()
            X_test_tensor = torch.from_numpy(X_test).float()
            self.shap_explainer[target_name] = shap.GradientExplainer(
                model.to(self.device), X_train_tensor.to(self.device))
            shap_values = self.shap_explainer[target_name](
                X_test_tensor.to(self.device)).values
        elif self.explainer == "explainer":
            self.shap_explainer[target_name] = shap.Explainer(
                model.predict, X_train)
            explanation = self.shap_explainer[target_name](X_test, silent=True)
            shap_values = explanation.values
        else:
            raise ValueError(
                f"Unknown explainer: {self.explainer}. "
                f"Please select one of: kernel, gradient, explainer.")

        return shap_values

    def _add_zeroes(self, target, correlated_features):
        features = [f for f in self.feature_names if f != target]
        for correlated_feature in correlated_features:
            correlated_feature_position = features.index(correlated_feature)
            self.all_mean_shap_values[-1] = np.insert(
                self.all_mean_shap_values[-1], correlated_feature_position, 0.)

    def predict(
            self,
            X,
            root_causes=None,
            prior: Optional[List[List[str]]] = None) -> nx.DiGraph:
        """
        Builds a causal graph from the shap values using a selection mechanism based
        on clustering, knee or abrupt methods.

        Parameters
        ----------
        X : pd.DataFrame
            The input data. Consists of all the features in a pandas DataFrame.
        root_causes : List[str], optional
            The root causes of the graph. If None, all features are considered as
            root causes, by default None.
        prior : List[List[str]], optional
            The prior knowledge about the connections between the features. If None,
            all features are considered as valid candidates for the connections, by
            default None.

        Returns
        -------
        nx.DiGraph
            The causal graph.
        """
        if self.verbose:
            print("-----\nshap.predict()")

        if not self.is_fitted_:
            raise ValueError("This Rex instance is not fitted yet. \
                Call 'fit' with appropriate arguments before using this estimator.")
        self.prior = prior

        # Recompute mean_shap_percentile here, in case it was changed
        self._compute_scaled_shap_threshold()

        # Who is calling me?
        caller_name = self._get_method_caller_name()

        if self.prog_bar and (not self.verbose):
            pbar_name = f"({caller_name}) SHAP_predict"
            pbar = ProgBar().start_subtask(pbar_name, 4 + len(self.feature_names))
        else:
            pbar = None

        # Reshape SHAP values if necessary
        for feature in self.feature_names:
            if self.shap_values[feature].shape[-1] == 1:
                self.shap_values[feature] = self.shap_values[feature]\
                    .reshape(self.shap_values[feature].shape[:-1])

        # Compute error contribution at this stage, since it needs the individual
        # SHAP values
        self.compute_error_contribution()
        pbar.update_subtask(pbar_name, 1) if pbar else None     # type: ignore

        self._compute_discrepancies(self.X_test)                # type: ignore
        pbar.update_subtask(pbar_name, 2) if pbar else None     # type: ignore

        self.connections = {}
        for target_idx, target in enumerate(self.feature_names):
            # The valid parents are the features that are in the same level of the
            # hierarchy as the target, or in previous levels. In case prior is not
            # provided, all features are valid candidates.
            candidate_causes = utils.valid_candidates_from_prior(
                self.feature_names, target, self.prior)

            print(
                f"Selecting features for target {target}...") if self.verbose else None

            feature_names_wo_target = [
                f for f in self.feature_names if f != target]

            # Select the features that are connected to the target
            self.connections[target] = select_features(
                values=self.shap_values[target],
                feature_names=feature_names_wo_target,
                min_impact=self.min_impact,
                exhaustive=self.exhaustive,
                threshold=float(self.mean_shap_threshold),
                verbose=self.verbose)
            pbar.update_subtask(
                pbar_name, target_idx + 3) if pbar else None     # type: ignore

        dag = utils.digraph_from_connected_features(
            X, self.feature_names, self.models, self.connections, root_causes, prior,
            reciprocity=self.reciprocity, anm_iterations=self.iters,
            verbose=self.verbose)
        pbar.update_subtask(pbar_name, # type: ignore
            len( self.feature_names) + 3) if pbar else None

        dag = utils.break_cycles_if_present(
            dag, self.shap_discrepancies,                       # type: ignore
            self.prior, verbose=self.verbose)
        pbar.update_subtask(pbar_name,                          # type: ignore
            len(self.feature_names) + 4) if pbar else None

        pbar.remove(pbar_name) if pbar else None                # type: ignore

        return dag

    def adjust(
            self,
            graph: nx.DiGraph,
            increase_tolerance: float = 0.0,
            sd_upper: float = 0.1):

        # self._compute_shap_discrepancies(X)
        new_graph = self._adjust_edges_from_shap_discrepancies(
            graph, increase_tolerance, sd_upper)
        return new_graph

    def _compute_discrepancies(self, X: pd.DataFrame):
        """
        Compute the discrepancies between the SHAP values and the target values
        for all features and all targets.

        Parameters
        ----------
        X : pd.DataFrame
            The input data. Consists of all the features in a pandas DataFrame.

        Returns
        -------
        pd.DataFrame
            A dataframe containing the discrepancies for all features and all targets.
        """
        if not self.is_fitted_:
            raise ValueError("This Rex instance is not fitted yet. \
                Call 'fit' with appropriate arguments before using this estimator.")
        self.discrepancies = pd.DataFrame(columns=self.feature_names)
        self.shap_discrepancies = defaultdict(dict)
        for target_name in self.feature_names:

            X_features = X.drop(target_name, axis=1)
            y = X[target_name].values

            feature_names = [
                f for f in self.feature_names if f != target_name]

            self.discrepancies.loc[target_name] = 0
            self.shap_discrepancies[target_name] = defaultdict(ShapDiscrepancy)

            # Loop through all features and compute the discrepancy
            for parent_name in feature_names:
                # Take the data that is needed at this iteration
                parent_data = X_features[parent_name].values
                parent_pos = feature_names.index(parent_name)
                shap_data = self.shap_values[target_name][:, parent_pos]

                # Form three vectors to compute the discrepancy
                x = parent_data.reshape(-1, 1)
                s = shap_data.reshape(-1, 1)

                # Compute the discrepancy
                self.shap_discrepancies[target_name][parent_name] = \
                    self._compute_discrepancy(
                        x, y, s,
                        target_name,
                        parent_name)
                SD = self.shap_discrepancies[target_name][parent_name]
                self.discrepancies.loc[target_name,
                                       parent_name] = SD.shap_discrepancy

        return self.discrepancies

    def _compute_discrepancy(self, x, y, s, target_name, parent_name) -> ShapDiscrepancy:
        if isinstance(x, (pd.Series, pd.DataFrame)):
            x = x.values
        elif not isinstance(x, np.ndarray):
            x = np.array(x)
        if isinstance(y, (pd.Series, pd.DataFrame)):
            y = y.values
        elif not isinstance(y, np.ndarray):
            y = np.array(y)

        X = sm.add_constant(x)
        model_y = sm.OLS(y, X).fit()
        model_s = sm.OLS(s, X).fit()

        # Heteroskedasticity tests:
        # The null hypothesis (H0): Signifies that Homoscedasticity is present
        # The alternative hypothesis (H1): Signifies that Heteroscedasticity is present
        # If the p-value is less than the significance level (0.05), we reject the
        # null hypothesis and conclude that heteroscedasticity is present.
        try:
            test_shap = sms.het_breuschpagan(model_s.resid, X)
        except ValueError:
            test_shap = (0, 0)
        shap_heteroskedasticity = test_shap[1] < 0.05

        try:
            test_parent = sms.het_breuschpagan(model_y.resid, X)
        except ValueError:
            test_parent = (0, 0)
        parent_heteroskedasticity = test_parent[1] < 0.05

        corr = spearmanr(s, y)
        discrepancy = 1 - np.abs(corr.correlation)
        # The p-value is below 5%: we reject the null hypothesis that the two
        # distributions are the same, with 95% confidence.
        _, ks_pvalue = kstest(s[:, 0], y)

        return ShapDiscrepancy(
            target=target_name,
            parent=parent_name,
            shap_heteroskedasticity=shap_heteroskedasticity,
            parent_heteroskedasticity=parent_heteroskedasticity,
            shap_p_value=test_shap[1],
            parent_p_value=test_parent[1],
            shap_model=model_s,
            parent_model=model_y,
            shap_discrepancy=discrepancy,
            shap_correlation=corr.correlation,
            shap_gof=r2_score(y, s),
            ks_pvalue=ks_pvalue,
            ks_result="Equal" if ks_pvalue > 0.05 else "Different"
        )

    def _adjust_edges_from_shap_discrepancies(
            self,
            graph: nx.DiGraph,
            increase_tolerance: float = 0.0,
            sd_upper: float = 0.1):
        """
        Adjust the edges of the graph based on the discrepancy index. This method
        removes edges that have a discrepancy index larger than the given standard
        deviation tolerance. The method also removes edges that have a discrepancy
        index larger than the discrepancy index of the target.

        Parameters
        ----------
        graph : nx.DiGraph
            The graph to adjust the edges of.
        increase_tolerance : float, optional
            The tolerance for the increase in the discrepancy index, by default 0.0.
        sd_upper : float, optional
            The upper bound for the Shap Discrepancy, by default 0.1. This is the max
            difference between the SHAP Discrepancy in both causal directions.

        Returns
        -------
        networkx.DiGraph
            The graph with the edges adjusted.
        """

        new_graph = nx.DiGraph()
        new_graph.add_nodes_from(graph.nodes(data=True))
        new_graph.add_edges_from(graph.edges())
        edges_reversed = []

        # Experimental
        if self._increase_upper_tolerance(self.discrepancies):
            increase_upper_tolerance = True
        else:
            increase_upper_tolerance = False

        for target in self.feature_names:
            target_mean = np.mean(self.discrepancies.loc[target].values)
            # Experimental
            if increase_upper_tolerance:
                tolerance = target_mean * increase_tolerance
            else:
                tolerance = 0.0

            # Iterate over the features and check if the edge should be reversed.
            for feature in self.feature_names:
                # If the edge is already reversed, skip it.
                if (target, feature) in edges_reversed or \
                        feature == target or \
                        not new_graph.has_edge(feature, target):
                    continue

                forward_sd = self.discrepancies.loc[target, feature]
                reverse_sd = self.discrepancies.loc[feature, target]
                diff = np.abs(forward_sd - reverse_sd)
                vector = [forward_sd, reverse_sd, diff, 0., 0., 0., 0.]

                # If the forward standard deviation is less than the reverse
                # standard deviation and within the tolerance range, attempt
                # to reverse the edge.
                if (forward_sd < reverse_sd) and \
                        (forward_sd < (target_mean + tolerance)):
                    # If the difference between the standard deviations is within
                    # the acceptable range, reverse the edge and check for cycles
                    # in the graph.
                    if diff < sd_upper and diff > 0.002:
                        new_graph.remove_edge(feature, target)
                        new_graph.add_edge(target, feature)
                        edges_reversed.append((feature, target))
                        cycles = list(nx.simple_cycles(new_graph))
                        # If reversing the edge creates a cycle that includes both
                        # the target and feature nodes, reverse the edge back to
                        # its original direction and log the decision as discarded.
                        if len(cycles) > 0 and \
                                self._nodes_in_cycles(cycles, feature, target):
                            new_graph.remove_edge(target, feature)
                            edges_reversed.remove((feature, target))
                            new_graph.add_edge(feature, target)
                            self._debugmsg("Discarded(cycles)",
                                           target, target_mean, feature, tolerance,
                                           vector, sd_upper, cycles)
                        # If reversing the edge does not create a cycle, log the
                        # decision as reversed.
                        else:
                            self._debugmsg("(*) Reversed edge",
                                           target, target_mean, feature, tolerance,
                                           vector, sd_upper, cycles)
                    # If the difference between the standard deviations is not
                    # within the acceptable range, log the decision as discarded.
                    else:
                        self._debugmsg("Outside tolerance",
                                       target, target_mean, feature, tolerance,
                                       vector, sd_upper, [])
                # If the forward standard deviation is greater than the reverse
                # standard deviation and within the tolerance range, log the
                # decision as ignored.
                else:
                    self._debugmsg("Ignored edge",
                                   target, target_mean, feature, tolerance,
                                   vector, sd_upper, [])

        return new_graph

    def _nodes_in_cycles(self, cycles, feature, target) -> bool:
        """
        Check if the given nodes are in any of the cycles.
        """
        for cycle in cycles:
            if feature in cycle and target in cycle:
                return True
        return False

    def _increase_upper_tolerance(self, discrepancies: pd.DataFrame):
        """
        Increase the upper tolerance if the discrepancy matrix properties are
        suspicious. We found these suspicious values empirically in the polymoial case.
        """
        D = discrepancies.values
        D = np.nan_to_num(D)
        det = np.linalg.det(D)
        norm = np.linalg.norm(D)
        cond = np.linalg.cond(D)
        m1 = "(*)" if det < -0.5 else "   "
        m2 = "(*)" if norm > .7 else "   "
        m3 = "(*)" if cond > 1500 else "   "
        if self.verbose:
            print(f"    {m2}{norm=:.2f} & ({m1}{det=:.2f} | {m3}{cond=:.2f})")
        if norm > 7.0 and (det < -0.5 or cond > 2000):
            return True
        return False

    def _input_vector(self, discrepancies, target, feature, target_mean):
        """
        Builds a vector with the values computed from the discrepancy index.
        Used to feed the model in _adjust_from_model method.
        """
        source_mean = np.mean(discrepancies.loc[feature].values)
        forward_sd = discrepancies.loc[target, feature]
        reverse_sd = discrepancies.loc[feature, target]
        diff = np.abs(forward_sd - reverse_sd)
        sdiff = np.abs(forward_sd + reverse_sd)
        d1 = diff / sdiff
        d2 = diff / forward_sd

        # Build the input for the model
        input_vector = np.array(
            [forward_sd, reverse_sd, diff, d1, d2,
             target_mean, source_mean])

        return input_vector

    def _adjust_predictions_shape(self, predictions, target_shape):
        # Concatenate if predictions is a list
        if isinstance(predictions, list):
            predictions = np.concatenate(predictions)
        else:
            predictions = np.array(predictions)

        # Reshape if necessary
        if predictions.shape != target_shape:
            # Flatten
            predictions = predictions.reshape(target_shape)

        return predictions

    def compute_error_contribution(self):
        """
        Computes the error contribution of each feature for each target.
        If this value is positive, then it means that, on average, the presence of
        the feature in the model leads to a higher error. Thus, without that feature,
        the prediction would have been generally better. In other words, the feature
        is making more harm than good!
        On the contrary, the more negative this value, the more beneficial
        the feature is for the predictions since its presence leads to smaller errors.

        Returns:
        --------
        err_contrib: pd.DataFrame
            Error contribution of each feature for each target.
        """
        error_contribution = dict()
        predictions = self.models.predict(self.X_test)   # type: ignore

        predictions = self._adjust_predictions_shape(
            predictions, (self.X_test.shape[0], self.X_test.shape[1]))

        y_hat = pd.DataFrame(predictions, columns=self.feature_names)
        y_true = self.X_test
        for target in self.feature_names:
            shap_values = pd.DataFrame(
                self.shap_values[target],
                columns=[c for c in self.feature_names if c != target])
            error_contribution[target] = self._individual_error_contribution(
                shap_values, y_true[target], y_hat[target])
            # Add a 0.0 value at index target in the error contribution series
            error_contribution[target] = pd.concat([
                error_contribution[target],
                pd.Series({target: 0.0})
            ])
            # Sort the series by index
            error_contribution[target] = error_contribution[target].sort_index()

        self.error_contribution = pd.DataFrame(error_contribution)
        return self.error_contribution

    def _individual_error_contribution(self, shap_values, y_true, y_pred):
        """
        Compute the error contribution of each feature.
        If this value is positive, then it means that, on average, the presence of
        the feature in the model leads to a higher error. Thus, without that feature,
        the prediction would have been generally better. In other words, the feature
        is making more harm than good!
        On the contrary, the more negative this value, the more beneficial the
        feature is for the predictions since its presence leads to smaller errors.

        Parameters:
        -----------
        shap_values: pd.DataFrame
            Shap values for a given target.
        y_true: pd.Series
            Ground truth values for a given target.
        y_pred: pd.Series
            Predicted values for a given target.

        Returns:
        --------
        error_contribution: pd.Series
            Error contribution of each feature.
        """
        abs_error = (y_true - y_pred).abs()
        y_pred_wo_feature = shap_values.apply(lambda feature: y_pred - feature)
        abs_error_wo_feature = y_pred_wo_feature.apply(
            lambda feature: (y_true-feature).abs())
        error_diff = abs_error_wo_feature.apply(
            lambda feature: abs_error - feature)
        ind_error_contribution = error_diff.mean()

        return ind_error_contribution

    def _debugmsg(
            self,
            msg,
            target,
            target_threshold,
            feature,
            tolerance,
            vector,
            sd_upper,
            cycles):
        if not self.verbose:
            return
        forward_sd, reverse_sd, diff, _, _, _, _ = vector
        fwd_bwd = f"{GREEN}<{RESET}" if forward_sd < reverse_sd else f"{RED}≮{RESET}"
        fwd_tgt = f"{GREEN}<{RESET}" if forward_sd < target_threshold + \
            tolerance else f"{RED}>{RESET}"
        diff_upper = f"{GREEN}<{RESET}" if diff < sd_upper else f"{RED}>{RESET}"
        print(f" -  {msg:<17s}: {feature} -> {target}",
              f"fwd|bwd({forward_sd:.3f}{fwd_bwd}{reverse_sd:.3f});",
              f"fwd{fwd_tgt}⍴({target_threshold:.3f}+{tolerance:.2f});",
              f"𝛿({diff:.3f}){diff_upper}Up({sd_upper:.2f}); "
              # f"𝛿({diff:.3f}){diff_tol}tol({sd_tol:.2f});",
              f"µ:{forward_sd/target_threshold:.3f}"
              )
        if len(cycles) > 0 and self._nodes_in_cycles(cycles, feature, target):
            print(f"    ~~ Cycles: {cycles}")

    def _get_method_caller_name(self):
        """
        Determine the name of the method that called it. It does this by using
        the inspect module to get the outermost frame of the call stack and then
        extracting the name of the third frame. If the name is either __call__ or
        _run_step, it returns "ReX". If any exception occurs during this
        process, it returns "unknown".
        """
        try:
            curframe = inspect.currentframe()
            calframe = inspect.getouterframes(curframe, 2)
            caller_name = calframe[2][3]
            if caller_name == "__call__" or caller_name == "_run_step":
                caller_name = "ReX"
        except Exception:                               # pylint: disable=broad-except
            caller_name = "unknown"
        return caller_name

    def _plot_shap_summary(
            self,
            target_name: str,
            ax,
            max_features_to_display: int = 20,
            **kwargs):
        """
        Plots the summary of the SHAP values for a given target.

        Arguments:
        ----------
            feature_names: List[str]
                The names of the features.
            mean_shap_values: np.array
                The mean SHAP values for each feature.
            feature_inds: np.array
                The indices of the features to be plotted.
            selected_features: List[str]
                The names of the selected features.
            ax: Axis
                The axis in which to plot the summary. If None, a new figure is created.
            **kwargs: Dict
                Additional arguments to be passed to the plot.
        """

        figsize_ = kwargs.get('figsize', (6, 3))
        fig = None
        if ax is None:
            fig, ax = plt.subplots(1, 1, figsize=figsize_)

        feature_inds = self.feature_order[target_name][:max_features_to_display]
        if self.correlation_th is not None:
            feature_names = [f for f in self.feature_names if (f != target_name) & (
                f not in self.correlated_features[target_name])]
        else:
            feature_names = [
                f for f in self.feature_names if f != target_name]
        selected_features = list(self.connections[target_name])

        y_pos = np.arange(len(feature_inds))
        ax.grid(True, axis='x')
        ax.barh(y_pos, self.shap_mean_values[target_name][feature_inds],
                0.7, align='center', color="#0e73fa", alpha=0.8)
        ax.xaxis.set_major_formatter(FormatStrFormatter('%.3g'))
        ax.set_yticks(y_pos, [feature_names[i] for i in feature_inds])
        ax.set_xlabel("Avg. SHAP value")
        ax.set_title(
            target_name + r' $\leftarrow$ ' +
            (','.join(selected_features) if selected_features else 'ø'))

        # Recompute mean_shap_percentile here, in case it was changed
        self._compute_scaled_shap_threshold()

        xlims = ax.get_xlim()
        if xlims[1] < self.mean_shap_threshold:
            ax.set_xlim(right=self.mean_shap_threshold +
                        ((xlims[1] - xlims[0]) / 20))

        ax.axvline(x=self.mean_shap_threshold, color='red', linestyle='--',
                   linewidth=0.5)

        fig = ax.figure if fig is None else fig
        return fig

    def _plot_discrepancies(self, target_name: str, threshold: float = 10.0, **kwargs):
        """
        Plot the discrepancies between the target variable and each feature.

        Parameters
        ----------
        - target_name (str)
            The name of the target variable.
        - threshold (float)
            The threshold to use for selecting what features to plot. Only
            features with a discrepancy index below this threshold will be plotted.

        Returns:
            None
        """
        pass


def custom_main(exp_name,
                path="/Users/renero/phd/data/RC4/",
                output_path="/Users/renero/phd/output/RC4/",
                scale=False):
    """
    Runs a custom main function for the given experiment name.

    Args:
        experiment_name (str): The name of the experiment to run.
        path (str): The path to the data files.
        output_path (str): The path to the output files.

    Returns:
        None
    """

    ref_graph = utils.graph_from_dot_file(f"{path}{exp_name}.dot")
    data = pd.read_csv(f"{path}{exp_name}.csv")
    if scale:
        scaler = StandardScaler()
        data = pd.DataFrame(scaler.fit_transform(data), columns=data.columns)
    # Split the dataframe into train and test
    train = data.sample(frac=0.9, random_state=42)
    test = data.drop(train.index)

    rex = utils.load_experiment(f"{exp_name}_nn", output_path)
    rex.is_fitted_ = True
    print(f"Loaded experiment {exp_name}")

    rex.shaps = ShapEstimator(
        explainer="gradient",
        models=rex.models,
        correlation_th=0.5,
        mean_shap_percentile=0.8,
        iters=20,
        reciprocity=True,
        min_impact=1e-06,
        on_gpu=False,
        verbose=False,
        prog_bar=True,
        silent=False)
    rex.shaps.fit(train)
    rex.shaps.predict(test, rex.root_causes)


def sachs_main():
    experiment_name = "sachs_long"
    path = "/Users/renero/phd/data/RC3/"
    output_path = "/Users/renero/phd/output/RC3/"

    ref_graph = utils.graph_from_dot_file(f"{path}{experiment_name}.dot")
    data = pd.read_csv(f"{path}{experiment_name}.csv")
    scaler = StandardScaler()
    data = pd.DataFrame(scaler.fit_transform(data), columns=data.columns)

    # Split the dataframe into train and test
    train = data.sample(frac=0.9, random_state=42)
    test = data.drop(train.index)

    rex = utils.load_experiment(f"{experiment_name}_gbt", output_path)
    rex.is_fitted_ = True
    rex.shaps.is_fitted_ = True
    print(f"Loaded experiment {experiment_name}")

    rex.shaps.prog_bar = False
    rex.shaps.verbose = True
    rex.shaps.iters = 100
    rex.shaps.predict(test, rex.root_causes)
    print("fininshed")


if __name__ == "__main__":
    custom_main('toy_dataset')
